from cad_viewer_widget import AnimationTrack
from jupyter_cadquery.viewer import animate as viewer_animate


class Animation:
    def __init__(self, viewer=None):
        self.tracks = []
        self.viewer = viewer
        self.reset()

    def add_track(self, path, action, times, values):
        if path[0] != "/":
            path = f"/{path}"
        if self.viewer is None:
            self.tracks.append(AnimationTrack(path, action, times, values))
        else:
            self.viewer.add_track(AnimationTrack(path, action, times, values))

    def animate(self, speed=1):
        if self.viewer is None:
            viewer_animate(self.tracks, speed=speed)
        else:
            self.viewer.animate(speed=speed)

    def reset(self):
        if self.viewer is not None:
            self.viewer.clear_tracks()



#
# Copyright 2019 Bernhard Walter
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from ..utils import warn

warn(
    "jupyter_cadquery.cadquery is deprecated, please use jupyter_cadquery directly for import "
    + "and jupyter_cadquery.replay for replay functions",
    DeprecationWarning,
    "once",
)

from ..cad_objects import (
    Assembly,
    PartGroup,
    Part,
    Faces,
    Edges,
    Vertices,
    show,
    web_color,
)
from ..tools import auto_show, show_accuracy, show_constraints

from ..replay import replay, enable_replay, disable_replay, reset_replay



import html

import numpy as np
import cadquery as cq

from jupyter_cadquery import Part, PartGroup, Faces, Edges, Vertices, show
from jupyter_cadquery.cad_objects import to_assembly
from jupyter_cadquery.base import _tessellate_group
from .utils import numpy_to_json

# pylint: disable=protected-access
# pylint: disable=unnecessary-lambda
def auto_show():
    PartGroup._ipython_display_ = lambda self: self.show()
    Part._ipython_display_ = lambda self: self.show()
    Faces._ipython_display_ = lambda self: self.show(grid=None, axes=False)
    Edges._ipython_display_ = lambda self: self.show(grid=None, axes=False)
    Vertices._ipython_display_ = lambda self: self.show(grid=None, axes=False)

    print("Overwriting auto display for cadquery Workplane and Shape")

    try:
        del cq.Workplane._repr_html_  # pylint: disable=no-member
        del cq.Shape._repr_html_  # pylint: disable=no-member
    except:  # pylint: disable=bare-except
        pass
    cq.Workplane._ipython_display_ = lambda cad_obj: show(cad_obj)
    cq.Shape._ipython_display_ = lambda cad_obj: show(cad_obj)
    cq.Assembly._ipython_display_ = lambda cad_obj: show(cad_obj)
    cq.Sketch._ipython_display_ = lambda cad_obj: show(cad_obj)


def show_constraints(assy, qs):
    colors = [
        "#e41a1c",
        "#377eb8",
        "#4daf4a",
        "#984ea3",
        "#ff7f00",
        "#ffff33",
        "#a65628",
        "#f781bf",
        "#999999",
        "#8dd3c7",
        "#ffffb3",
        "#bebada",
        "#fb8072",
        "#80b1d3",
        "#fdb462",
        "#b3de69",
        "#fccde5",
        "#d9d9d9",
    ]

    constraints = []
    objects = []
    cache = {}

    for i, q1q2 in enumerate(qs):
        parts = []

        kind = q1q2[-1]

        if len(q1q2) == 3:
            q1q2 = ((q1q2[0].split("@")[0], q1q2[0]), (q1q2[1].split("@")[0], q1q2[1]))
        else:
            q1q2 = (q1q2[0:2], q1q2[2:4])

        for q in q1q2:
            name, shape = q
            if name in cache:
                obj = cache[name]["obj"]
                loc = cache[name]["loc"]
            else:
                obj = assy.objects[name].obj
                loc = assy.objects[name].loc

                parent = assy.objects[name].parent
                while parent is not None:
                    loc = parent.loc * loc
                    parent = parent.parent

                cache[name] = {"obj": obj, "loc": loc, "shape": shape}

                objects.append(Part(cq.Workplane(obj.val().located(loc)), name=name, show_faces=False))

            label = str(shape)
            if isinstance(shape, str):
                shape = assy._query(shape)[1]

            parts.append(
                Faces(
                    cq.Workplane(cq.Workplane(shape).val().located(loc)),
                    name=html.escape(label),
                    color=colors[i % len(colors)],
                )
            )
        constraints.append(PartGroup(parts, "%s_%d" % (kind, i)))

    show(PartGroup([PartGroup(objects, "objects")] + constraints), axes=True, axes0=True)


def show_accuracy(assy, cs):
    def relocate(name, shape):
        a = assy.objects[name]
        loc = a.loc

        parent = a.parent
        while parent is not None:
            loc = parent.loc * loc
            parent = parent.parent

        if isinstance(shape, str):
            shape = assy._query(shape)[1]
        return cq.Workplane(cq.Workplane(shape).val().located(loc))

    def center(face):
        c = face.Center()
        return np.array((c.x, c.y, c.z))

    def normal(face):
        n = face.normalAt()
        return np.array((n.x, n.y, n.z))

    def print_metric(results):
        l = max([len(r[1]) for r in results])
        h = ("Constraint", "Normal-Dist", "Normal-Angle", "Point-Dist")
        print(f"{h[0]:{l+7}s} {h[1]:12s}  {h[2]:12s}  {h[3]:12s}")
        print("-" * (l + 46))
        for kind, label, nrm_dist, nrm_angle, pnt_dist in results:
            metric = f"{kind:5s} {label:{l}s} "
            metric += " " * 27 if nrm_dist is None else f"{nrm_dist:12.9f}  {nrm_angle:12.8}°"
            metric += " " * 13 if pnt_dist is None else f"{pnt_dist:12.9f}"
            print(metric)

    results = []
    for q1q2 in cs:
        kind = q1q2[-1]

        if len(q1q2) == 3:
            n_q1q2 = ((q1q2[0].split("@")[0], q1q2[0]), (q1q2[1].split("@")[0], q1q2[1]))
            label = "%s - %s" % q1q2[:2]
        else:
            n_q1q2 = (q1q2[0:2], q1q2[2:4])
            label = "%s<%s> - %s<%s>" % (q1q2[0], q1q2[1].__class__.__name__, q1q2[2], q1q2[3].__class__.__name__)

        shape1 = relocate(*n_q1q2[0])
        shape2 = relocate(*n_q1q2[1])

        pnt_dist = None
        nrm_dist = None
        nrm_angle = None

        if kind in ["Point", "Plane"]:
            c1, c2 = center(shape1.val()), center(shape2.val())
            pnt_dist = np.linalg.norm(c1 - c2)
        if kind in ["Axis", "Plane"]:
            n1, n2 = normal(shape1.val()), normal(shape2.val())
            nrm_dist = np.linalg.norm(n1 + n2)  # distance between n1 and -n2 since n1 and n2 point opposite
            c = np.dot(n1, -n2) / np.linalg.norm(n1) / np.linalg.norm(n2)
            nrm_angle = np.arccos(np.clip(c, -1, 1)) / np.pi * 180

        results.append((kind, label, nrm_dist, nrm_angle, pnt_dist))

    print_metric(results)


def cq_to_json(obj, indent=None):
    shapes, states = _tessellate_group(to_assembly(obj), {}, None, False)
    return [numpy_to_json(shapes, indent=indent), states]



import json

from jupyter_cadquery.tessellator import cache, make_key
import multiprocessing
from multiprocessing import shared_memory
from cachetools import cached
from jupyter_cadquery.mp_tess import mp_tess
from jupyter_cadquery.ocp_utils import serialize

pool = None


class KeyMapper:
    def __init__(self):
        self.counter = 0
        self.map = {}

    def reset(self):
        self.counter = 0
        self.map = {}

    def add(self, key):
        path = f"obj{self.counter}"
        self.counter += 1
        self.map[path] = key
        return path

    def get_key(self, path):
        return self.map.get(path)


keymap = KeyMapper()


def clear_shared_mem(path):
    try:
        s = shared_memory.SharedMemory(path)
        s.close()
        s.unlink()
    except:  # pylint: disable=bare-except
        ...


def init_pool():
    global pool  # pylint: disable=global-statement
    if pool is None:
        pool = multiprocessing.Pool(int(multiprocessing.cpu_count() * 0.8))


def close_pool():
    global pool  # pylint: disable=global-statement
    pool.close()
    pool.join()
    pool = None


def get_mp_result(apply_result):
    path, result = apply_result.get()

    # update cache to hold full result instead of ApplyResult object
    key = keymap.get_key(path)
    cache.__setitem__(key, result)

    clear_shared_mem(path)
    return result


def is_apply_result(obj):
    return isinstance(obj, multiprocessing.pool.ApplyResult)


# This will cache the ApplyResult object after calling mp_tess in an async way.
# Cache will be updated in get_mp_result so that it holds the actual result
@cached(cache, key=make_key)
def mp_tessellate(
    shapes,
    deviation,  # only provided for managing cache
    quality,
    angular_tolerance,
    compute_faces=True,
    compute_edges=True,
    debug=False,
):

    shape = shapes[0]

    key = make_key(shape, deviation, quality, angular_tolerance, compute_edges=True, compute_faces=True)
    path = keymap.add(key)

    clear_shared_mem(path)

    s = serialize(shape)
    sm = shared_memory.SharedMemory(path, True, len(s))
    sm.buf[: len(s)] = s

    result = pool.apply_async(
        mp_tess, (path, deviation, quality, angular_tolerance, compute_faces, compute_edges, debug)
    )

    return result



#
# Copyright 2019 Bernhard Walter
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np

from cadquery import Compound, __version__

from cad_viewer_widget import show as viewer_show, get_default_sidecar, _set_default_sidecar, get_sidecar

from jupyter_cadquery.progress import Progress
from jupyter_cadquery.utils import Color, Timer, warn
from jupyter_cadquery.ocp_utils import bounding_box, get_point, loc_to_tq, BoundingBox, wrapped_or_None, np_bbox
from jupyter_cadquery.tessellator import discretize_edge, tessellate, compute_quality, bbox_edges
from jupyter_cadquery.mp_tessellator import (
    is_apply_result,
    mp_tessellate,
    get_mp_result,
    keymap,
    init_pool,
    close_pool,
)
from jupyter_cadquery.defaults import (
    get_default,
    apply_defaults,
    create_args,
    add_shape_args,
    get_defaults,
    tessellation_args,
    show_args,
    preset,
)


UNSELECTED = 0
SELECTED = 1
EMPTY = 3

#
# Simple Part and PartGroup classes
#


class _CADObject(object):
    def __init__(self):
        self.color = Color(get_default("default_color"))

    def to_state(self):
        raise NotImplementedError("not implemented yet")

    def collect_shapes(
        self,
        path,
        loc,
        deviation,
        angular_tolerance,
        edge_accuracy,
        render_edges,
        parallel,
        progress,
        timeit,
    ):
        raise NotImplementedError("not implemented yet")


class _Part(_CADObject):
    def __init__(self, shape, name="Part", color=None, show_faces=True, show_edges=True):
        super().__init__()
        self.name = name
        self.id = None
        self.color = Color(get_default("default_color") if color is None else color)

        self.shape = shape
        self.set_states(show_faces, show_edges)
        self.renderback = False

    def set_states(self, show_faces, show_edges):
        self.state_faces = SELECTED if show_faces else UNSELECTED
        self.state_edges = SELECTED if show_edges else UNSELECTED

    def to_state(self):
        return [self.state_faces, self.state_edges]

    def collect_shapes(
        self,
        path,
        loc,
        deviation,
        angular_tolerance,
        edge_accuracy,
        render_edges,
        parallel=False,
        progress=None,
        timeit=False,
    ):
        self.id = f"{path}/{self.name}"

        # A first rough estimate of the bounding box.
        # Will be too large, but is sufficient for computing the quality
        with Timer(timeit, self.name, "compute quality:", 2) as t:
            bb = bounding_box(self.shape, loc=wrapped_or_None(loc), optimal=False)
            quality = compute_quality(bb, deviation=deviation)
            t.info = str(bb)

        with Timer(timeit, self.name, "tessellate:     ", 2) as t:
            func = mp_tessellate if parallel else tessellate
            result = func(
                self.shape,
                deviation=deviation,
                quality=quality,
                angular_tolerance=angular_tolerance,
                debug=timeit,
                compute_edges=render_edges,
            )

            t.info = f"{{quality:{quality:.4f}, angular_tolerance:{angular_tolerance:.2f}}}"

            t, q = loc_to_tq(wrapped_or_None(loc))
            if parallel and is_apply_result(result):
                mesh = {"result": result, "t": t, "q": q}
                bb = {}
            else:
                mesh = result
                bb = np_bbox(result["vertices"], t, q)

        if progress is not None:
            progress.update()

        if isinstance(self.color, tuple):
            color = [c.web_color for c in self.color]  # pylint: disable=not-an-iterable
            alpha = 1.0
        else:
            color = self.color.web_color
            alpha = self.color.a

        return {
            "id": self.id,
            "type": "shapes",
            "name": self.name,
            "shape": mesh,
            "color": color,
            "alpha": alpha,
            "renderback": self.renderback,
            "accuracy": quality,
            "bb": bb,
        }

    def compound(self):
        return Compound._makeCompound(self.shape)

    def compounds(self):
        return [self.compound()]


class _Faces(_Part):
    def __init__(self, faces, name="Faces", color=None, show_faces=True, show_edges=True):
        super().__init__(faces, name, color, show_faces, show_edges)
        self.color = Color(color or (238, 130, 238))
        self.renderback = True


class _Edges(_CADObject):
    def __init__(self, edges, name="Edges", color=None, width=1):
        super().__init__()
        self.shape = edges
        self.name = name
        self.id = None

        if color is not None:
            if isinstance(color, (list, tuple)) and isinstance(color[0], Color):
                self.color = color
            elif isinstance(color, Color):
                self.color = color
            else:
                self.color = Color(color)
        self.width = width

    def to_state(self):
        return [EMPTY, SELECTED]

    def collect_shapes(
        self,
        path,
        loc,
        deviation,
        angular_tolerance,
        edge_accuracy,
        render_edges,
        parallel=False,
        progress=None,
        timeit=False,
    ):
        self.id = f"{path}/{self.name}"

        with Timer(timeit, self.name, "bounding box:", 2) as t:
            bb = bounding_box(self.shape, loc=wrapped_or_None(loc))
            quality = compute_quality(bb, deviation=deviation)
            deflection = quality / 100 if edge_accuracy is None else edge_accuracy
            t.info = str(bb)

        with Timer(timeit, self.name, "discretize:  ", 2):
            edges = []
            for edge in self.shape:
                edges.extend(discretize_edge(edge, deflection))
            edges = np.asarray(edges)

        if progress:
            progress.update()

        color = [c.web_color for c in self.color] if isinstance(self.color, tuple) else self.color.web_color

        return {
            "id": self.id,
            "type": "edges",
            "name": self.name,
            "shape": edges,
            "color": color,
            "width": self.width,
            "bb": bb.to_dict(),
        }


class _Vertices(_CADObject):
    def __init__(self, vertices, name="Vertices", color=None, size=1):
        super().__init__()
        self.shape = vertices
        self.name = name
        self.id = None
        self.color = Color(color or (148, 0, 211))
        self.size = size

    def to_state(self):
        return [EMPTY, SELECTED]

    def collect_shapes(
        self,
        path,
        loc,
        deviation,
        angular_tolerance,
        edge_accuracy,
        render_edges,
        parallel=False,
        progress=None,
        timeit=False,
    ):
        self.id = f"{path}/{self.name}"

        bb = bounding_box(self.shape, loc=wrapped_or_None(loc))

        if progress is not None:
            progress.update()

        return {
            "id": self.id,
            "type": "vertices",
            "name": self.name,
            "shape": [get_point(vertex) for vertex in self.shape],
            "color": self.color.web_color,
            "size": self.size,
            "bb": bb.to_dict(),
        }


class _PartGroup(_CADObject):
    def __init__(self, objects, name="Group", loc=None):
        super().__init__()
        self.objects = objects
        self.name = name
        self.loc = loc
        self.id = None

    def to_nav_dict(self):
        return {
            "type": "node",
            "name": self.name,
            "id": self.id,
            "children": [obj.to_nav_dict() for obj in self.objects],
        }

    def collect_shapes(
        self,
        path,
        loc,
        deviation,
        angular_tolerance,
        edge_accuracy,
        render_edges,
        parallel=False,
        progress=None,
        timeit=False,
    ):

        self.id = f"{path}/{self.name}"

        if loc is None and self.loc is None:
            combined_loc = None
        elif loc is None:
            combined_loc = self.loc
        else:
            combined_loc = loc * self.loc

        result = {
            "parts": [],
            "loc": None if self.loc is None else loc_to_tq(wrapped_or_None(self.loc)),
            "name": self.name,
        }
        for obj in self.objects:
            result["parts"].append(
                obj.collect_shapes(
                    self.id,
                    combined_loc,
                    deviation,
                    angular_tolerance,
                    edge_accuracy,
                    render_edges,
                    parallel,
                    progress,
                    timeit,
                )
            )
        return result

    def to_state(self, parents=None):  # pylint: disable=arguments-differ
        parents = parents or ()
        result = {}
        for i, obj in enumerate(self.objects):
            if isinstance(obj, _PartGroup):
                for k, v in obj.to_state((*parents, i)).items():
                    result[k] = v
            else:
                result[str(obj.id)] = obj.to_state()
        return result

    def count_shapes(self):
        def c(pg):
            count = 0
            for p in pg.objects:
                if isinstance(p, _PartGroup):
                    count += c(p)
                else:
                    count += 1
            return count

        return c(self)

    def compounds(self):
        result = []
        for obj in self.objects:
            result += obj.compounds()
        return result

    def compound(self):
        return Compound._makeCompound(self.compounds())  # pylint: disable=protected-access


def _tessellate_group(group, kwargs=None, progress=None, timeit=False):
    if kwargs is None:
        kwargs = {}

    shapes = group.collect_shapes(
        "",
        None,
        deviation=preset("deviation", kwargs.get("deviation")),
        angular_tolerance=preset("angular_tolerance", kwargs.get("angular_tolerance")),
        edge_accuracy=preset("edge_accuracy", kwargs.get("edge_accuracy")),
        render_edges=preset("render_edges", kwargs.get("render_edges")),
        parallel=kwargs.get("parallel"),
        progress=progress,
        timeit=timeit,
    )
    states = group.to_state()

    return shapes, states


def _combined_bb(shapes):
    def c_bb(shapes, bb):
        for shape in shapes["parts"]:
            if shape.get("parts") is None:
                if bb is None:
                    if shape["bb"] is None:
                        bb = BoundingBox()
                    else:
                        bb = BoundingBox(shape["bb"])
                else:
                    if shape["bb"] is not None:
                        bb.update(shape["bb"])

                # after updating the global bounding box, remove the local
                del shape["bb"]
            else:
                bb = c_bb(shape, bb)
        return bb

    bb = c_bb(shapes, None)
    return bb


def mp_get_results(shapes, progress):
    def walk(shapes):
        for shape in shapes["parts"]:
            if shape.get("parts") is None:
                if shape.get("type") == "shapes":
                    if is_apply_result(shape["shape"].get("result")):
                        mesh = get_mp_result(shape["shape"]["result"])
                        t = shape["shape"]["t"]
                        q = shape["shape"]["q"]
                        shape["shape"] = mesh
                        shape["bb"] = np_bbox(mesh["vertices"], t, q)

                    if progress is not None:
                        progress.update()
            else:
                walk(shape)

    walk(shapes)
    return shapes


def get_accuracies(shapes):
    def _get_accuracies(shapes, lengths):
        if shapes.get("parts"):
            for shape in shapes["parts"]:
                _get_accuracies(shape, lengths)
        elif shapes.get("type") == "shapes":
            accuracies[shapes["id"]] = shapes["accuracy"]

    accuracies = {}
    _get_accuracies(shapes, accuracies)
    return accuracies


def get_normal_len(render_normals, shapes, deviation):
    if render_normals:
        accuracies = get_accuracies(shapes)
        normal_len = max(accuracies.values()) / deviation * 4
    else:
        normal_len = 0

    return normal_len


def insert_bbox(bbox, shapes, states):
    # derive the top level states path part
    prefix = list(states)[0].split("/")[1]

    bbox = {
        "id": f"/{prefix}/BoundingBox",
        "type": "edges",
        "name": "BoundingBox",
        "shape": bbox_edges(bbox),
        "color": "#FF00FF",
        "width": 1,
        "bb": bbox,
    }
    # inject bounding box into shapes
    shapes["parts"].insert(0, bbox)
    # and states
    states[f"/{prefix}/BoundingBox"] = [3, 1]


def _show(part_group, **kwargs):
    for k in kwargs:
        if get_default(k, "n/a") == "n/a":
            raise KeyError(f"Paramater {k} is not a valid argument for show()")

    if kwargs.get("cad_width") is not None and kwargs.get("cad_width") < 640:
        warn("cad_width has to be >= 640, setting to 640")
        kwargs["cad_width"] = 640

    if kwargs.get("height") is not None and kwargs.get("height") < 400:
        warn("height has to be >= 400, setting to 400")
        kwargs["height"] = 400

    if kwargs.get("tree_width") is not None and kwargs.get("tree_width") < 250:
        warn("tree_width has to be >= 250, setting to 250")
        kwargs["tree_width"] = 250

    if kwargs.get("quality") is not None:
        warn("quality is ignored. Use deviation to control smoothness of edges")
        del kwargs["quality"]

    sidecar_backup = None

    # if kwargs.get("parallel") is not None:
    #     if kwargs["parallel"] and platform.system() != "Linux":
    #         warn("parallel=True only works on Linux. Setting parallel=False")
    #         kwargs["parallel"] = False

    timeit = preset("timeit", kwargs.get("timeit"))

    with Timer(timeit, "", "overall"):

        if part_group is None:

            import base64  # pylint:disable=import-outside-toplevel
            import pickle  # pylint:disable=import-outside-toplevel
            from jupyter_cadquery.logo import LOGO_DATA  # pylint:disable=import-outside-toplevel

            logo = pickle.loads(base64.b64decode(LOGO_DATA))

            config = add_shape_args(logo["config"])

            defaults = get_defaults()
            config["cad_width"] = defaults["cad_width"]
            config["tree_width"] = defaults["tree_width"]
            config["height"] = defaults["height"]
            config["glass"] = defaults["glass"]
            config["title"] = defaults["viewer"]

            for k, v in create_args(kwargs).items():
                config[k] = v

            shapes = logo["data"]["shapes"]
            states = logo["data"]["states"]
            bb = _combined_bb(shapes).to_dict()
            # add global bounding box
            shapes["bb"] = bb

        else:

            config = apply_defaults(**kwargs)

            if config.get("viewer") == "":
                # If viewer is "" (the show default), then the default sidecar should be taken into account
                config["viewer"] = None
                viewer = get_sidecar()
            elif config.get("viewer") is None:
                # if viewer is None (explicitely set), then ignore the default sidecar, i.e. back it up and set to None
                sidecar_backup = get_default_sidecar()
                _set_default_sidecar(None)
                viewer = None
            else:
                viewer = get_sidecar(config["viewer"])

            if viewer is not None:
                # Clear remaining animation tracks. They might not fit to the next assembly
                viewer.clear_tracks()

            if config.get("reset_camera") is False:  #  could be None
                if config.get("zoom") is not None:
                    del config["zoom"]
                if config.get("position") is not None:
                    del config["position"]
                if config.get("quaternion") is not None:
                    del config["quaternion"]

            parallel = preset("parallel", config.get("parallel"))
            with Timer(timeit, "", "tessellate", 1):
                num_shapes = part_group.count_shapes()
                progress_len = 2 * num_shapes if parallel else num_shapes
                progress = None if num_shapes < 2 else Progress(progress_len)

                if parallel:
                    init_pool()
                    keymap.reset()

                shapes, states = _tessellate_group(part_group, tessellation_args(config), progress, timeit)

                if parallel:
                    mp_get_results(shapes, progress)
                    close_pool()

                bb = _combined_bb(shapes).to_dict()
                # add global bounding box
                shapes["bb"] = bb

                if progress is not None:
                    progress.done()

            # Calculate normal length

            config["normal_len"] = get_normal_len(
                preset("render_normals", config.get("render_normals")),
                shapes,
                preset("deviation", config.get("deviation")),
            )

            show_bbox = preset("show_bbox", kwargs.get("show_bbox"))
            if show_bbox:
                insert_bbox(show_bbox, shapes, states)

        with Timer(timeit, "", "show shapes", 1):
            cv = viewer_show(shapes, states, **show_args(config))

            # If we forced to ignore the default sidecar, restore it
            if sidecar_backup is not None:
                _set_default_sidecar(sidecar_backup)

    return cv



#
# Copyright 2019 Bernhard Walter
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import cadquery as cq
from .cad_objects import _PartGroup, _Part
from .ocp_utils import write_stl_file, is_compound


def exportSTL(cadObj, filename, tolerance=None, angular_tolerance=None):
    compound = None
    if isinstance(cadObj, (_PartGroup, _Part)):
        compound = cadObj.compound()
    elif is_compound(cadObj):
        compound = cadObj
    elif isinstance(cadObj, (cq.Shape, cq.Workplane)):
        compound = _Part(cadObj).compound()
    else:
        print("Unsupported CAD object %s, convert to PartGroup or Part" % type(cadObj))

    if compound is not None:
        write_stl_file(
            compound,
            filename,
            tolerance=tolerance,
            angular_tolerance=angular_tolerance,
        )



import pickle

import cadquery as cq

from jupyter_cadquery.ocp_utils import loc_to_tq, tq_to_loc, serialize, deserialize
from jupyter_cadquery.base import _tessellate_group, _combined_bb
from jupyter_cadquery.cad_objects import to_assembly
from jupyter_cadquery.utils import numpy_to_json

try:
    from cadquery_massembly import MAssembly, Mate
    from cadquery_massembly.massembly import MateDef

    HAS_MASSEMBLY = True
except ImportError:
    HAS_MASSEMBLY = False


def save_binary(assembly, filename, metadata=None):
    def to_tuple(vec):
        return vec.toTuple() if isinstance(vec, cq.Vector) else vec

    def _create_assembly_object(name, loc=None, color=None, shape=None, children=None):
        return {
            "name": name,
            "loc": loc,
            "color": color,
            "shape": shape,
            "children": [] if children is None else children,
        }

    def _save_binary(assembly):
        if assembly is None:
            return None

        result = _create_assembly_object(
            assembly.name,
            None if assembly.loc is None else loc_to_tq(assembly.loc.wrapped),
            None if assembly.color is None else assembly.color.toTuple(),
            None if assembly.obj is None else serialize(assembly.obj.val().wrapped),
            [_save_binary(child) for child in assembly.children],
        )

        return result

    typ = type(assembly).__name__
    mates = {}
    if typ == "MAssembly":
        mates = {
            name: {
                "mate": {
                    "origin": to_tuple(mate_def.mate.origin),
                    "x_dir": to_tuple(mate_def.mate.x_dir),
                    "y_dir": to_tuple(mate_def.mate.y_dir),
                    "z_dir": to_tuple(mate_def.mate.z_dir),
                },
                "assembly": mate_def.assembly.name,
                "origin": mate_def.origin,
            }
            for name, mate_def in assembly.mates.items()
        }

    objs = {
        "type": typ,
        "obj": _save_binary(assembly),
        "mates": mates,
        "metadata": metadata,
    }
    with open(filename, "wb") as fd:
        pickle.dump(objs, fd)


def load_binary(filename, assembly_name=None):
    def to_workplane(obj):
        return cq.Workplane(obj=cq.Solid(obj))

    def _load_binary(obj, klass):
        if obj is None:
            return None

        assembly = klass(
            None if obj["shape"] is None else to_workplane(deserialize(obj["shape"])),
            name=obj["name"],
            loc=None if obj["loc"] is None else cq.Location(tq_to_loc(*obj["loc"])),
            color=None if obj["color"] is None else cq.Color(*obj["color"]),
        )
        for child in obj["children"]:
            assembly.add(_load_binary(child, klass))

        return assembly

    with open(filename, "rb") as fd:
        buffer = pickle.load(fd)

    is_ma = buffer["type"] == "MAssembly"

    klass = MAssembly if is_ma else cq.Assembly
    assembly = _load_binary(buffer["obj"], klass)

    if is_ma:
        mates = {}
        for name, mate_obj in buffer["mates"].items():
            mate = Mate()
            mate.origin = cq.Vector(*mate_obj["mate"]["origin"])
            mate.x_dir = cq.Vector(*mate_obj["mate"]["x_dir"])
            mate.y_dir = cq.Vector(*mate_obj["mate"]["y_dir"])
            mate.z_dir = cq.Vector(*mate_obj["mate"]["z_dir"])
            mates[name] = MateDef(mate, assembly.objects[mate_obj["assembly"]], mate_obj["origin"])
        assembly.mates = mates

    if assembly_name is not None:
        assembly.name = assembly_name

    return assembly, buffer["metadata"]


def exportJson(cad_obj, filename):
    shapes, states = _tessellate_group(to_assembly(cad_obj))
    bb = _combined_bb(shapes).to_dict()
    # add global bounding box
    shapes["bb"] = bb

    with open(filename, "w", encoding="utf-8") as fd:
        fd.write(numpy_to_json((shapes, states)))



#
# Copyright 2019 Bernhard Walter
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import warnings

from cad_viewer_widget import (
    open_viewer as cvw_open_viewer,
    AnimationTrack,
)

from cad_viewer_widget.sidecar import (
    get_sidecar as get_viewer,
    get_sidecars as get_viewers,
    close_sidecar as close_viewer,
    close_sidecars as close_viewers,
    set_default as set_default_viewer,
)

from cad_viewer_widget._version import __version__ as cvw_version
from ._version import __version_info__ as jcq_version_info, __version__ as jcq_version
from .ocp_utils import occt_version
from .stepreader import StepReader

from .cad_objects import (
    Assembly,
    PartGroup,
    Part,
    Faces,
    Edges,
    Vertices,
    show,
    web_color,
    get_pick,
)

from .defaults import (
    get_default,
    get_defaults,
    set_defaults,
    reset_defaults,
    create_args,
)

from .utils import warn
from .tools import auto_show


def versions():
    print("jupyter_cadquery ", jcq_version)
    print("cad_viewer_widget", cvw_version)
    print("open cascade     ", occt_version())


def open_viewer(viewer=None, default=True, **kwargs):
    args = create_args(get_defaults())
    for k, v in kwargs.items():
        args[k] = v
    args["title"] = viewer
    cv = cvw_open_viewer(**args)

    set_defaults(reset_camera=True)

    if kwargs.get("cad_width") is not None:
        set_defaults(cad_width=kwargs["cad_width"])
    if kwargs.get("tree_width") is not None:
        set_defaults(tree_width=kwargs["tree_width"])
    if kwargs.get("height") is not None:
        set_defaults(height=kwargs["height"])
    if kwargs.get("theme") is not None:
        set_defaults(theme=kwargs["theme"])
    if kwargs.get("glass") is not None:
        set_defaults(glass=kwargs["glass"])

    if default:
        set_default_viewer(viewer)

    show(viewer=viewer, **kwargs)
    return cv


def set_sidecar(title, anchor="right", init=False):
    warn(
        "set_sidecar(title, init=False) is deprecated, please use: open_viewer(title='CadQuery', **kwargs)",
        DeprecationWarning,
        "once",
    )
    if init:
        with warnings.catch_warnings(record=True):
            warnings.simplefilter("ignore", DeprecationWarning)
            open_viewer(viewer=title, default=True, anchor=anchor)


def close_sidecar(title):
    warn(
        "close_sidecar(title) is deprecated, please use close_viewer(title)",
        DeprecationWarning,
        "once",
    )

    close_viewer(title)


def close_sidecars():
    warn(
        "close_sidecars() is deprecated, please use close_viewers()",
        DeprecationWarning,
        "once",
    )

    close_viewers()


try:
    from IPython import get_ipython

    shell_name = get_ipython().__class__.__name__
    if shell_name == "ZMQInteractiveShell":
        auto_show()
except Exception as ex:
    ...



import json
import math
import numpy as np
import time
import warnings
from webcolors import name_to_rgb, hex_to_rgb, rgb_to_hex


def round_sig(x, sig):
    return round(x, sig - int(math.floor(math.log10(abs(x)))) - 1)


class Color:
    def __init__(self, color=None):
        self.a = 1.0
        if color is None:
            self.r = self.g = self.b = 160
        elif isinstance(color, Color):
            self.r, self.g, self.b, self.a = color.r, color.g, color.b, color.a
        elif isinstance(color, str):
            if color[0] == "#":
                if len(color) > 7:
                    c = hex_to_rgb(color[:7])
                    self.a = int(color[7:9], 16) / 255
                else:
                    c = hex_to_rgb(color)
            else:
                c = name_to_rgb(color)
            self.r = c.red
            self.g = c.green
            self.b = c.blue
        elif isinstance(color, (tuple, list)) and len(color) >= 3:
            if all((isinstance(c, float) and (c <= 1.0) and (c >= 0.0)) for c in color[:3]):
                self.r, self.g, self.b = (int(c * 255) for c in color[:3])
            elif all((isinstance(c, int) and (c <= 255) and (c >= 0)) for c in color[:3]):
                self.r, self.g, self.b = color[:3]
            else:
                self._invalid(color)

            if len(color) == 4:
                self.a = color[3] if color[3] <= 1.0 else color[3] / 100
        else:
            self._invalid(color)

    def __str__(self):
        return f"Color({self.r}, {self.g}, {self.b}, {self.a})"

    def __repr__(self):
        return f"Color(({self.r}, {self.g}, {self.b}, {self.a}))"

    def _invalid(self, color):
        print(f"warning: {color} is an invalid color, using grey (#aaa)")
        self.r = self.g = self.b = 160
        self.a = 1.0

    @property
    def rgb(self):
        return (self.r, self.g, self.b)

    @property
    def rgba(self):
        return (self.r, self.g, self.b, self.a)

    @property
    def percentage(self):
        return (self.r / 255, self.g / 255, self.b / 255)

    @property
    def web_color(self):
        return rgb_to_hex((self.r, self.g, self.b))


#
# Helpers
#


def explode(edge_list):
    return [[edge_list[i], edge_list[i + 1]] for i in range(len(edge_list) - 1)]


def flatten(nested_list):
    return [y for x in nested_list for y in x]


def numpy_to_json(obj, indent=None):
    class NumpyArrayEncoder(json.JSONEncoder):
        def default(self, o):
            if isinstance(o, np.integer):
                return int(o)
            if isinstance(o, np.floating):
                return float(o)
            if isinstance(o, np.ndarray):
                return o.tolist()

            return super(NumpyArrayEncoder, self).default(o)

    return json.dumps(obj, cls=NumpyArrayEncoder, indent=indent)


def distance(v1, v2):
    return np.linalg.norm([x - y for x, y in zip(v1, v2)])


#
# tree search
#


def tree_find_single_selector(tree, selector):
    if tree.name == selector:
        return tree

    for c in tree.children:
        result = tree_find_single_selector(c, selector)
        if result is not None:
            return result
    return None


class Timer:
    def __init__(self, timeit, name, activity, level=0):
        if isinstance(timeit, bool):
            self.timeit = 99 if timeit else -1
        else:
            self.timeit = timeit
        self.activity = activity
        self.name = name
        self.level = level
        self.info = ""
        self.start = time.time()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if self.level <= self.timeit:
            prefix = ""
            if self.level > 0:
                prefix += "| " * self.level

            if self.name != "":
                self.name = f'"{self.name}"'

            print("%8.3f sec: %s%s %s %s" % (time.time() - self.start, prefix, self.activity, self.name, self.info))


def px(w):
    return f"{w}px"


def warn(message, warning=RuntimeWarning, when="always"):
    def warning_on_one_line(
        message, category, filename, lineno, file=None, line=None
    ):  # pylint: disable=unused-argument
        return "%s: %s" % (category.__name__, message)

    warn_format = warnings.formatwarning
    warnings.formatwarning = warning_on_one_line
    warnings.simplefilter(when, warning)
    warnings.warn(message + "\n", warning)
    warnings.formatwarning = warn_format
    warnings.simplefilter("ignore", warning)



#
# Copyright 2019 Bernhard Walter
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

try:
    from cadquery_massembly import MAssembly, Mate

    HAS_MASSEMBLY = True
except ImportError:
    HAS_MASSEMBLY = False

from cadquery.occ_impl.shapes import Face, Edge, Wire
from cadquery import (
    Workplane,
    Sketch,
    Shape,
    Compound,
    Vector,
    Vertex,
    Location,
    Assembly as CqAssembly,
    Color as CqColor,
)

from jupyter_cadquery.base import _PartGroup, _Part, _Edges, _Faces, _Vertices, _show, _tessellate_group, _combined_bb

from .utils import Color, flatten, warn, numpy_to_json
from .ocp_utils import get_rgba, is_compound, is_shape
from .defaults import get_default, preset


EDGE_COLOR = "Silver"
THICK_EDGE_COLOR = "MediumOrchid"
VERTEX_COLOR = "MediumOrchid"
FACE_COLOR = "Violet"


def web_color(name):
    wc = Color(name)
    return CqColor(*wc.percentage)


class Part(_Part):
    def __init__(self, shape, name="Part", color=None, show_faces=True, show_edges=True):
        if color is None:
            color = get_default("default_color")
        self.cq_shape = shape
        super().__init__(_to_occ(shape), name, color, show_faces, show_edges)

    def to_assembly(self):
        return PartGroup([self])

    def show(self, grid=None, axes=False):
        if grid is None:
            grid = [False, False, False]
        return show(self, grid=grid, axes=axes)


class Faces(_Faces):
    def __init__(self, faces, name="Faces", color=None, show_faces=True, show_edges=True):
        self.cq_shape = faces
        super().__init__(_to_occ(faces.combine()), name, color, show_faces, show_edges)

    def to_assembly(self):
        return PartGroup([self])

    def show(self, grid=None, axes=False):
        if grid is None:
            grid = [False, False, False]
        return show(self, grid=grid, axes=axes)


class Edges(_Edges):
    def __init__(self, edges, name="Edges", color=None, width=1):
        self.cq_shape = edges
        super().__init__(_to_occ(edges), name, color, width)

    def to_assembly(self):
        return PartGroup([self])

    def show(self, grid=None, axes=False):
        if grid is None:
            grid = [False, False, False]
        return show(self, grid=grid, axes=axes)


class Vertices(_Vertices):
    def __init__(self, vertices, name="Vertices", color=None, size=1):
        self.cq_shape = vertices
        super().__init__(_to_occ(vertices), name, color, size)

    def to_assembly(self):
        return PartGroup([self])

    def show(self, grid=None, axes=False):
        if grid is None:
            grid = [False, False, False]
        return show(self, grid=grid, axes=axes)


class PartGroup(_PartGroup):
    def to_assembly(self):
        return self

    def show(self, grid=None, axes=False):
        if grid is None:
            grid = [False, False, False]
        return show(self, grid=grid, axes=axes)

    def add(self, cad_obj):
        self.objects.append(cad_obj)

    def add_list(self, cad_objs):
        self.objects += cad_objs

    def get_pick(self, pick):
        if pick == {}:
            print("First double click on an object in the CAD viewer")
        else:
            objs = [o for o in self.objects if o.id == f'{pick["path"]}/{pick["name"]}']
            if objs:
                return objs[0].cq_shape
            else:
                print(f"no object found for pick {pick}")
        return None


def get_pick(assembly, pick):
    if pick == {}:
        print("First double click on an object in the CAD viewer")
        return None
    if isinstance(assembly, PartGroup):
        return assembly.get_pick(pick)
    else:
        path = pick["path"]
        name = pick["name"]
        id_ = "/".join([path, name])

        short_path = "/".join(id_.split("/")[2:])
        if assembly.objects.get(short_path) is not None:
            return assembly.objects[short_path]
        else:
            short_path = "/".join(id_.split("/")[2:-1])
            if assembly.objects.get(short_path) is not None:
                return assembly.objects[short_path]
        return None


class Assembly(PartGroup):
    def __init__(self, *args, **kwargs):
        import warnings

        super().__init__(*args, **kwargs)
        warnings.warn(
            "Class 'Assembly' is deprecated (too many assemblies ...). Please use class 'PartGroup' instead",
            RuntimeWarning,
        )


def _to_occ(cad_obj):
    def sketch_to_occ(sketch):
        locs = sketch.locs if sketch.locs else [Location()]
        if sketch._faces:  # pylint:disable=protected-access
            objs = flatten([sketch._faces.moved(loc).Faces() for loc in locs])  # pylint:disable=protected-access
        else:
            objs = [edge.moved(loc) for edge in sketch._edges for loc in locs]  # pylint:disable=protected-access

        return [obj.wrapped for obj in objs]

    # special case Wire, must be handled before Workplane
    if _is_wirelist(cad_obj) or _is_edgelist(cad_obj):
        all_edges = []
        for edges in cad_obj.objects:
            all_edges += edges.Edges()
        return [edge.wrapped for edge in all_edges]

    elif isinstance(cad_obj, Sketch):
        return sketch_to_occ(cad_obj)

    elif isinstance(cad_obj, Workplane):
        result = []
        for obj in cad_obj.objects:
            if isinstance(obj, Sketch):
                result += sketch_to_occ(obj)
            else:
                result.append(obj.wrapped)
        return result

    elif isinstance(cad_obj, Shape):
        return [cad_obj.wrapped]

    elif is_compound(cad_obj):
        return [cad_obj]

    elif is_shape(cad_obj):
        return [cad_obj]

    else:
        raise NotImplementedError(type(cad_obj))


def _parent(cad_obj, obj_id):
    if cad_obj.parent is not None:
        if isinstance(cad_obj.parent.val(), Vector):
            return _from_vectorlist(
                cad_obj.parent,
                obj_id,
                name="Parent",
                color=Color(EDGE_COLOR),
                show_parent=False,
            )
        elif isinstance(cad_obj.parent.val(), Vertex):
            return _from_vertexlist(
                cad_obj.parent,
                obj_id,
                name="Parent",
                color=Color(EDGE_COLOR),
                show_parent=False,
            )
        elif isinstance(cad_obj.parent.val(), Edge):
            return _from_edgelist(
                cad_obj.parent,
                obj_id,
                name="Parent",
                color=Color(EDGE_COLOR),
                show_parent=False,
            )
        elif isinstance(cad_obj.parent.val(), Wire):
            return [_from_wirelist(cad_obj.parent, obj_id, name="Parent", color=Color(EDGE_COLOR))]
        else:
            return [
                Part(
                    cad_obj.parent,
                    "Parent_%d" % obj_id,
                    show_edges=True,
                    show_faces=False,
                )
            ]
    else:
        return []


def _from_facelist(cad_obj, obj_id, name="Faces", show_parent=True):
    result = [Faces(cad_obj, "%s_%d" % (name, obj_id), color=Color(FACE_COLOR))]
    if show_parent:
        result = _parent(cad_obj, obj_id) + result
    return result


def _from_edgelist(cad_obj, obj_id, name="Edges", color=None, show_parent=True):
    result = [Edges(cad_obj, "%s_%d" % (name, obj_id), color=Color(color or THICK_EDGE_COLOR), width=3)]
    if show_parent:
        result = _parent(cad_obj, obj_id) + result
    return result


def _from_wirelist(cad_obj, obj_id, name="Edges", color=None, show_parent=True):
    result = [Edges(cad_obj, "%s_%d" % (name, obj_id), color=Color(color or THICK_EDGE_COLOR), width=3)]
    if show_parent:
        result = _parent(cad_obj, obj_id) + result
    return result


def _from_vector(vec, obj_id, name="Vector", show_parent=True):
    tmp = Workplane()
    obj = tmp.newObject([vec])
    return _from_vectorlist(obj, obj_id, name, show_parent=show_parent)


def _from_vectorlist(cad_obj, obj_id, name="Vertices", color=None, show_parent=True):
    if cad_obj.vals():
        vectors = cad_obj.vals()
    else:
        vectors = [cad_obj.val()]
    obj = cad_obj.newObject([Vertex.makeVertex(v.x, v.y, v.z) for v in vectors])
    result = [Vertices(obj, "%s_%d" % (name, obj_id), color=Color(color or VERTEX_COLOR), size=6)]
    if show_parent:
        result = _parent(cad_obj, obj_id) + result
    return result


def _from_vertexlist(cad_obj, obj_id, name="Vertices", color=None, show_parent=True):
    result = [Vertices(cad_obj, "%s_%d" % (name, obj_id), color=Color(color or VERTEX_COLOR), size=6)]
    if show_parent:
        result = _parent(cad_obj, obj_id) + result
    return result


# pylint:disable=protected-access
def _from_sketch(cad_obj, obj_id, show_parent=True, show_selection=True):

    result = []

    locs = cad_obj.locs if cad_obj.locs else [Location()]

    workplane = Workplane()
    if cad_obj._faces:
        for loc in locs:
            workplane.objects += cad_obj._faces.moved(loc).Faces()
        result += _from_facelist(workplane, obj_id, name="Faces", show_parent=show_parent)
    elif cad_obj._edges:
        workplane.objects = [edge.moved(loc) for edge in cad_obj._edges for loc in locs]
        result += _from_edgelist(workplane, obj_id, name="Edges", show_parent=show_parent)

    if show_selection and cad_obj._selection:
        workplane = Workplane()
        if isinstance(cad_obj._selection[0], Location):
            workplane.objects = [
                Vertex.makeVertex(0, 0, 0).moved(loc * obj) for obj in cad_obj._selection for loc in locs
            ]
            sel = _from_vertexlist(workplane, obj_id, name="Locations", show_parent=show_parent)

        elif isinstance(cad_obj._selection[0], Face):
            for loc in locs:
                workplane.objects += flatten([obj._faces.moved(loc).Faces() for obj in cad_obj._selection])
            sel = _from_facelist(workplane, obj_id, name="Faces", show_parent=show_parent)

        elif isinstance(cad_obj._selection[0], (Edge, Wire)):
            workplane.objects = [edge.moved(loc) for edge in cad_obj._selection for loc in locs]
            sel = _from_edgelist(workplane, obj_id, name="Edges", show_parent=show_parent)

        elif isinstance(cad_obj._selection[0], Vertex):
            workplane.objects = [vertex.moved(loc) for vertex in cad_obj._selection for loc in locs]
            sel = _from_vertexlist(workplane, obj_id, name="Vertices", show_parent=show_parent)

        result.append(PartGroup(sel, name=f"Selection_{obj_id}"))

    return result


def to_edge(mate, loc=None, scale=1) -> Workplane:
    w = Workplane()
    for d in (mate.x_dir, mate.y_dir, mate.z_dir):
        edge = Edge.makeLine(mate.origin, mate.origin + d * scale)
        w.objects.append(edge if loc is None else edge.moved(loc))

    return w


def _from_mate(cad_obj, name="Mate", mate_scale=1):
    rgb = (Color((255, 0, 0)), Color((0, 128, 0)), Color((0, 0, 255)))
    return Edges(to_edge(cad_obj, scale=mate_scale), name=name, width=3, color=rgb)


def from_assembly(cad_obj, top, loc=None, render_mates=False, mate_scale=1, default_color=None):
    loc = Location()
    render_loc = cad_obj.loc

    if cad_obj.color is None:
        if default_color is None:
            color = Color(get_default("default_color"))
        else:
            color = Color(default_color)
    else:
        color = Color(get_rgba(cad_obj.color))

    # Special handling for edge lists in an MAssembly
    is_edges = [isinstance(obj, Edge) for obj in cad_obj.shapes]
    if is_edges and all(is_edges):
        if cad_obj.color is None:
            if default_color is None:
                color = Color(get_default("default_edgecolor"))
            else:
                color = Color(default_color)
        else:
            color = Color(get_rgba(cad_obj.color))

        workplane = Workplane()
        workplane.objects = cad_obj.shapes
        parent = [
            Edges(
                workplane,
                name="%s_0" % cad_obj.name,
                color=color,
            )
        ]
    else:
        if cad_obj.color is None:
            if default_color is None:
                color = Color(get_default("default_color"))
            else:
                color = Color(default_color)
        else:
            color = Color(get_rgba(cad_obj.color))
        parent = [
            Part(
                Workplane(shape),
                "%s_%d" % (cad_obj.name, i),
                color=color,
            )
            for i, shape in enumerate(cad_obj.shapes)
        ]

    if render_mates and cad_obj.mates is not None:
        rgb = (Color((255, 0, 0)), Color((0, 128, 0)), Color((0, 0, 255)))
        pg = PartGroup(
            [
                Edges(to_edge(mate_def.mate, scale=mate_scale), name=name, width=3, color=rgb)
                for name, mate_def in top.mates.items()
                if mate_def.assembly == cad_obj
            ],
            name="mates",
            loc=Location(),  # mates inherit the parent location, so actually add a no-op
        )
        if pg.objects:
            parent.append(pg)

    children = [from_assembly(c, top, loc, render_mates, mate_scale) for c in cad_obj.children]
    return PartGroup(parent + children, cad_obj.name, loc=render_loc)


def _from_workplane(cad_obj, obj_id, name="Part", default_color=None, show_parent=False):
    result = Part(cad_obj, "%s_%d" % (name, obj_id), color=Color(default_color))
    # if show_parent:
    #     result = _parent(cad_obj, obj_id) + result
    return result


def _is_facelist(cad_obj):
    return (
        hasattr(cad_obj, "objects")
        and cad_obj.objects != []
        and all([isinstance(obj, Face) for obj in cad_obj.objects])
    )


def _is_vertexlist(cad_obj):
    return (
        hasattr(cad_obj, "objects")
        and cad_obj.objects != []
        and all([isinstance(obj, Vertex) for obj in cad_obj.objects])
    )


def _is_edgelist(cad_obj):
    return (
        hasattr(cad_obj, "objects")
        and cad_obj.objects != []
        and all([isinstance(obj, Edge) for obj in cad_obj.objects])
    )


def _is_wirelist(cad_obj):
    return (
        hasattr(cad_obj, "objects")
        and cad_obj.objects != []
        and all([isinstance(obj, Wire) for obj in cad_obj.objects])
    )


def _debug(msg):
    # print("DEBUG:", msg)
    pass


def to_assembly(
    *cad_objs, names=None, name="Group", render_mates=None, mate_scale=1, default_color=None, show_parent=True
):
    if names is None:
        names = [None] * len(cad_objs)

    default_color = get_default("default_color") if default_color is None else default_color
    assembly = PartGroup([], name)
    obj_id = 0

    for obj_name, cad_obj in zip(names, cad_objs):
        if isinstance(cad_obj, (PartGroup, Part, Faces, Edges, Vertices)):
            _debug(f"CAD Obj {obj_id}: PartGroup, Part, Faces, Edges, Vertices")
            assembly.add(cad_obj)

        elif HAS_MASSEMBLY and isinstance(cad_obj, MAssembly):
            _debug(f"CAD Obj {obj_id}: MAssembly")
            assembly.add(
                from_assembly(
                    cad_obj, cad_obj, render_mates=render_mates, mate_scale=mate_scale, default_color=default_color
                )
            )

        elif HAS_MASSEMBLY and isinstance(cad_obj, Mate):
            _debug(f"CAD Obj {obj_id}: Mate")
            obj_name = "Mate" if obj_name is None else obj_name
            assembly.add(_from_mate(cad_obj, name=obj_name, mate_scale=mate_scale))

        elif isinstance(cad_obj, CqAssembly):
            _debug(f"CAD Obj {obj_id}: cqAssembly")
            assembly.add(from_assembly(cad_obj, cad_obj, default_color=default_color))

        elif isinstance(cad_obj, Sketch):
            _debug(f"CAD Obj {obj_id}: Sketch")
            assembly.add_list(
                _from_sketch(
                    cad_obj,
                    obj_id,
                    show_parent=show_parent,
                )
            )

        elif isinstance(cad_obj, Edge):
            _debug(f"CAD Obj {obj_id}: Edge")
            obj_name = "Faces" if obj_name is None else obj_name
            assembly.add_list(_from_edgelist(Workplane(cad_obj), obj_id, obj_name, show_parent=show_parent))

        elif isinstance(cad_obj, Face):
            _debug(f"CAD Obj {obj_id}: Face")
            obj_name = "Faces" if obj_name is None else obj_name
            assembly.add_list(_from_facelist(Workplane(cad_obj), obj_id, obj_name, show_parent=show_parent))

        elif isinstance(cad_obj, Wire):
            _debug(f"CAD Obj {obj_id}: Wire")
            obj_name = "Wires" if obj_name is None else obj_name
            assembly.add(_from_wirelist(Workplane(cad_obj), obj_id, obj_name, show_parent=show_parent))

        elif isinstance(cad_obj, Vertex):
            _debug(f"CAD Obj {obj_id}: Vertex")
            obj_name = "Vertices" if obj_name is None else obj_name
            assembly.add_list(_from_vertexlist(Workplane(cad_obj), obj_id, obj_name, show_parent=show_parent))

        elif _is_facelist(cad_obj):
            _debug(f"CAD Obj {obj_id}: facelist")
            obj_name = "Faces" if obj_name is None else obj_name
            assembly.add_list(_from_facelist(cad_obj, obj_id, obj_name, show_parent=show_parent))

        elif _is_edgelist(cad_obj):
            _debug(f"CAD Obj {obj_id}: edgelist")
            obj_name = "Edges" if obj_name is None else obj_name
            assembly.add_list(_from_edgelist(cad_obj, obj_id, obj_name, show_parent=show_parent))

        elif _is_wirelist(cad_obj):
            _debug(f"CAD Obj {obj_id}: wirelist")
            obj_name = "Wires" if obj_name is None else obj_name
            assembly.add_list(_from_wirelist(cad_obj, obj_id, obj_name, show_parent=show_parent))

        elif _is_vertexlist(cad_obj):
            _debug(f"CAD Obj {obj_id}: vertexlist")
            obj_name = "Vertices" if obj_name is None else obj_name
            assembly.add_list(_from_vertexlist(cad_obj, obj_id, obj_name, show_parent=show_parent))

        elif isinstance(cad_obj, Vector):
            _debug(f"CAD Obj {obj_id}: Vector")
            obj_name = "Vector" if obj_name is None else obj_name
            assembly.add_list(_from_vector(cad_obj, obj_id, obj_name, show_parent=show_parent))

        elif isinstance(cad_obj, (Shape, Compound)):
            _debug(f"CAD Obj {obj_id}: Shape, Compound")
            obj_name = "Part" if obj_name is None else obj_name
            assembly.add(
                _from_workplane(
                    Workplane(cad_obj), obj_id, obj_name, default_color=default_color, show_parent=show_parent
                )
            )

        elif is_compound(cad_obj):
            _debug(f"CAD Obj {obj_id}: TopoDS Compound")
            obj_name = "Compound" if obj_name is None else obj_name
            assembly.add(_Part([cad_obj], name=obj_name, color=default_color))

        elif is_shape(cad_obj):
            _debug(f"CAD Obj {obj_id}: TopoDS Shape")
            obj_name = "Shape" if obj_name is None else obj_name
            assembly.add(_Part([cad_obj], name=obj_name, color=default_color))

        elif isinstance(cad_obj.val(), Vector):
            _debug(f"CAD Obj {obj_id}: Vector val()")
            obj_name = "Vector" if obj_name is None else obj_name
            assembly.add_list(_from_vectorlist(cad_obj, obj_id, obj_name, show_parent=show_parent))

        elif isinstance(cad_obj, Workplane):
            _debug(f"CAD Obj {obj_id}: Workplane")
            obj_name = "Part" if obj_name is None else obj_name
            if len(cad_obj.vals()) == 1 and not isinstance(cad_obj.val(), Sketch):
                assembly.add(
                    _from_workplane(cad_obj, obj_id, obj_name, default_color=default_color, show_parent=show_parent)
                )
            else:
                assembly2 = PartGroup([], name="Group_%s" % obj_id)
                for j, obj in enumerate(cad_obj.vals()):
                    if isinstance(obj, Sketch):
                        for sketch_obj in _from_sketch(obj, j, show_parent=False):
                            assembly2.add(sketch_obj)
                    else:
                        assembly2.add(
                            _from_workplane(
                                Workplane(obj), j, obj_name, default_color=default_color, show_parent=show_parent
                            )
                        )
                assembly.add(assembly2)

        else:
            raise NotImplementedError("Type:", cad_obj)

        obj_id += 1
    return assembly


def show(*cad_objs, names=None, **kwargs):
    """Show CAD objects in Jupyter

    Valid keywords:

    DISPLAY OPTIONS
    - viewer:             Name of the sidecar viewer (default=""):
                          "" uses the default sidecar (if exists) and None forces to use notebook cell
    - anchor:             How to open sidecar: "right", "split-right", "split-bottom", ... (default="right")
    - cad_width:          Width of CAD view part of the view (default=800)
    - tree_width:         Width of navigation tree part of the view (default=250)
    - height:             Height of the CAD view (default=600)
    - theme:              Theme "light" or "dark" (default="light")
    - pinning:            Allow replacing the CAD View by a canvas screenshot (default=True in cells, else False)

    TESSELLATION OPTIONS
    - angular_tolerance:  Shapes: Angular deflection in radians for tessellation (default=0.2)
    - deviation:          Shapes: Deviation from linear deflection value (default=0.1)
    - edge_accuracy:      Edges: Precision of edge discretization (default=None, i.e. mesh quality / 100)
    - default_color:      Default face color (default=(232, 176, 36))
    - default_edge_color: Default edge color (default="#707070")
    - optimal_bb:         Use optimal bounding box (default=False)
    - render_normals:     Render vertex normals(default=False)
    - render_edges:       Render edges  (default=True)
    - render_mates:       Render mates (for MAssemblies, default=False)
    - mate_scale:         Scale of rendered mates (for MAssemblies, default=1)

    VIEWER OPTIONS
    - control:            Use trackball controls ('trackball') or orbit controls ('orbit') (default='trackball')
    - up:                 Use z-axis ('Z') or y-axis ('Z') as up direction for the camera
    - axes:               Show axes (default=False)
    - axes0:              Show axes at (0,0,0) (default=False)
    - grid:               Show grid (default=[False, False, False])
    - ticks:              Hint for the number of ticks in both directions (default=10)
    - ortho:              Use orthographic projections (default=True)
    - transparent:        Show objects transparent (default=False)
    - black_edges:        Show edges in black (default=False)
    - position:           Absolute camera position that will be scaled (default=None)
    - quaternion:         Camera rotation as quaternion (x, y, z, w) (default=None)
    - target:             Camera target to look at (default=None)
    - zoom:               Zoom factor of view (default=2.5)
    - reset_camera:       Reset camera position, rotation and zoom to default (default=True)
    - zoom_speed:         Mouse zoom speed (default=1.0)
    - pan_speed:          Mouse pan speed (default=1.0)
    - rotate_speed:       Mouse rotate speed (default=1.0)
    - ambient_intensity   Intensity of ambient light (default=0.75)
    - direct_intensity    Intensity of direct lights (default=0.15)
    - show_parent:        Show the parent for edges, faces and vertices objects
    - show_bbox:          Show bounding box (default=False)
    - collapse:           Collapse CAD tree (1: collapse nodes with single leaf, 2: collapse all nodes)
    - cad_width:          Width of CAD view part of the view (default=800)
    - tree_width:         Width of navigation tree part of the view (default=250)
    - height:             Height of the CAD view (default=600)
    - tools:              Show the viewer tools like the object tree (default=True)
    - glass:              Show the viewer in glass mode, i.e (CAD navigation as transparent overlay (default=False)
    - timeit:             Show rendering times, levels = False, 0,1,2,3,4,5 (default=False)
    - parallel:           (Linux only) Whether to use multiprocessing for parallel tessellation
    - js_debug:           Enable debug output in browser console (default=False)

    NOT SUPPORTED ANY MORE:
    - mac_scrollbar       The default now
    - bb_factor:          Removed
    - display             Use 'viewer="<viewer title>"' (for sidecar display) or 'viewer=None' (for cell display)
    - quality             Use 'deviation'to control smoothness of rendered edges
    """

    render_mates = preset("render_mates", kwargs.get("render_mates"))
    mate_scale = preset("mate_scale", kwargs.get("mate_scale"))
    default_color = preset("default_color", kwargs.get("default_color"))
    show_parent = preset("show_parent", kwargs.get("show_parent"))

    if isinstance(kwargs.get("grid"), bool):
        warn(
            "Using bool for grid is deprecated, please use (xy-grid, xz-grid. yz-grid)",
            DeprecationWarning,
            "once",
        )
        kwargs["grid"] = (kwargs["grid"], False, False)

    if cad_objs:

        assembly = to_assembly(
            *cad_objs,
            names=names,
            render_mates=render_mates,
            mate_scale=mate_scale,
            default_color=default_color,
            show_parent=show_parent,
        )

        if assembly is None:
            raise ValueError("%s cannot be viewed" % cad_objs)

        if len(assembly.objects) == 1 and isinstance(assembly.objects[0], PartGroup):
            # omit leading "PartGroup" group
            return _show(assembly.objects[0], **kwargs)
        else:
            return _show(assembly, **kwargs)

    else:

        return _show(None, **kwargs)



"""Tessellator class"""

import os
import sys

from cachetools import LRUCache, cached

import numpy as np

# pylint: disable=no-name-in-module,import-error
from OCP.gp import gp_Vec, gp_Pnt
from OCP.BRep import BRep_Tool
from OCP.BRepTools import BRepTools
from OCP.BRepGProp import BRepGProp_Face
from OCP.BRepMesh import BRepMesh_IncrementalMesh
from OCP.TopLoc import TopLoc_Location
from OCP.TopAbs import TopAbs_Orientation
from OCP.TopTools import TopTools_IndexedDataMapOfShapeListOfShape, TopTools_IndexedMapOfShape
from OCP.TopExp import TopExp, TopExp_Explorer
from OCP.TopAbs import TopAbs_EDGE, TopAbs_FACE, TopAbs_SOLID
from OCP.TopoDS import TopoDS
from OCP.BRepAdaptor import BRepAdaptor_Curve
from OCP.GCPnts import GCPnts_QuasiUniformDeflection


from cadquery.occ_impl.shapes import Compound
from jupyter_cadquery.utils import Timer, round_sig
from jupyter_cadquery.ocp_utils import get_faces, np_bbox, loc_to_tq

MAX_HASH_KEY = 2147483647


#
# Caching helpers
#


def make_key(
    shape, deviation, quality, angular_tolerance, compute_edges=True, compute_faces=True, debug=False
):  # pylint: disable=unused-argument
    # quality is a measure of bounding box and deviation, hence can be ignored (and should due to accuracy issues
    # of non optimal bounding boxes. debug and progress are also irrelevant for tessellation results)
    if not isinstance(shape, (tuple, list)):
        shape = [shape]

    key = (
        tuple((s.HashCode(MAX_HASH_KEY) for s in shape)),
        deviation,
        angular_tolerance,
        compute_edges,
        compute_faces,
    )
    return key


def get_size(obj):
    size = sys.getsizeof(obj)
    if isinstance(obj, dict):
        size += sum([get_size(v) + len(k) for k, v in obj.items()])
    elif isinstance(obj, np.ndarray):
        size += obj.size * obj.dtype.itemsize
    elif isinstance(obj, (tuple, list)):
        size += sum([get_size(i) for i in obj])
    return size


cache_size = os.environ.get("JCQ_CACHE_SIZE_MB")
if cache_size is None:
    cache_size = 128 * 1024 * 1024
else:
    cache_size = int(cache_size) * 1024 * 1024
cache = LRUCache(maxsize=cache_size, getsizeof=get_size)


class Tessellator:
    def __init__(self):
        self.vertices = np.empty((0, 3), dtype="float32")
        self.triangles = np.empty((0,), dtype="uint32")
        self.normals = np.empty((0, 3), dtype="float32")
        self.normals = np.empty((0, 2, 3), dtype="float32")
        self.shape = None
        self.edges = []

    def number_solids(self, shape):
        count = 0
        e = TopExp_Explorer(shape, TopAbs_SOLID)
        while e.More():
            count += 1
            e.Next()
        return count

    def compute(
        self,
        shape,
        quality,
        angular_tolerance,
        compute_faces=True,
        compute_edges=True,
        debug=False,
    ):

        self.shape = shape

        count = self.number_solids(shape)
        with Timer(debug, "", f"mesh incrementally {'(parallel)' if count > 1 else ''}", 3):
            # Remove previous mesh data
            BRepTools.Clean_s(shape)
            BRepMesh_IncrementalMesh(shape, quality, False, angular_tolerance, count > 1)

        if compute_faces:
            with Timer(debug, "", "get nodes, triangles and normals", 3):
                self.tessellate()

        if compute_edges:
            with Timer(debug, "", "get edges", 3):
                self.compute_edges()

        # Remove mesh data again
        # BRepTools.Clean_s(shape)

    def tessellate(self):
        self.vertices = []
        self.triangles = []
        self.normals = []

        # global buffers
        p_buf = gp_Pnt()
        n_buf = gp_Vec()
        loc_buf = TopLoc_Location()

        offset = -1

        # every line below is selected for performance. Do not introduce functions to "beautify" the code

        for face in get_faces(self.shape):
            if face.Orientation() == TopAbs_Orientation.TopAbs_REVERSED:
                i1, i2 = 2, 1
            else:
                i1, i2 = 1, 2

            internal = face.Orientation() == TopAbs_Orientation.TopAbs_INTERNAL

            poly = BRep_Tool.Triangulation_s(face, loc_buf)
            if poly is not None:
                Trsf = loc_buf.Transformation()

                # add vertices
                flat = []
                for i in range(1, poly.NbNodes() + 1):
                    flat.extend(poly.Node(i).Transformed(Trsf).Coord())
                self.vertices.extend(flat)

                # add triangles
                flat = []
                for i in range(1, poly.NbTriangles() + 1):
                    coord = poly.Triangle(i).Get()
                    flat.extend((coord[0] + offset, coord[i1] + offset, coord[i2] + offset))
                self.triangles.extend(flat)

                # add normals
                if poly.HasUVNodes():
                    prop = BRepGProp_Face(face)
                    flat = []
                    for i in range(1, poly.NbNodes() + 1):
                        u, v = poly.UVNode(i).Coord()
                        prop.Normal(u, v, p_buf, n_buf)
                        if n_buf.SquareMagnitude() > 0:
                            n_buf.Normalize()
                        flat.extend(n_buf.Reverse().Coord() if internal else n_buf.Coord())
                    self.normals.extend(flat)

                offset += poly.NbNodes()

    def compute_edges(self):
        edge_map = TopTools_IndexedMapOfShape()
        face_map = TopTools_IndexedDataMapOfShapeListOfShape()

        TopExp.MapShapes_s(self.shape, TopAbs_EDGE, edge_map)
        TopExp.MapShapesAndAncestors_s(self.shape, TopAbs_EDGE, TopAbs_FACE, face_map)

        for i in range(1, edge_map.Extent() + 1):
            edge = TopoDS.Edge_s(edge_map.FindKey(i))

            face_list = face_map.FindFromKey(edge)
            if face_list.Extent() == 0:
                # print("no faces")
                continue

            loc = TopLoc_Location()

            face = TopoDS.Face_s(face_list.First())
            triangle = BRep_Tool.Triangulation_s(face, loc)
            poly = BRep_Tool.PolygonOnTriangulation_s(edge, triangle, loc)

            if poly is None:
                continue

            if hasattr(poly, "Node"):  # OCCT > 7.5
                nrange = range(1, poly.NbNodes() + 1)
                index = poly.Node
            else:  # OCCT == 7.5
                indices = poly.Nodes()
                nrange = range(indices.Lower(), indices.Upper() + 1)
                index = indices.Value

            transf = loc.Transformation()
            v1 = None
            for j in nrange:
                v2 = triangle.Node(index(j)).Transformed(transf).Coord()
                if v1 is not None:
                    self.edges.append((v1, v2))
                v1 = v2

    def get_vertices(self):
        return np.asarray(self.vertices, dtype=np.float32)

    def get_triangles(self):
        return np.asarray(self.triangles, dtype=np.int32)

    def get_normals(self):
        return np.asarray(self.normals, dtype=np.float32)

    def get_edges(self):
        return np.asarray(self.edges, dtype=np.float32)


def compute_quality(bb, deviation=0.1):
    # Since tessellation caching depends on quality, try to come up with stable a quality value
    quality = round_sig(
        (round_sig(bb.xsize, 3) + round_sig(bb.ysize, 3) + round_sig(bb.zsize, 3)) / 300 * deviation, 3
    )
    return quality


# cache key: (shape.hash, deviaton, angular_tolerance, compute_edges, compute_faces)
@cached(cache, key=make_key)
def tessellate(
    shapes,
    # only provided for managing cache:
    deviation: float,  # pylint: disable=unused-argument
    quality: float,
    angular_tolerance: float,
    compute_faces=True,
    compute_edges=True,
    debug=False,
):
    compound = Compound._makeCompound(shapes) if len(shapes) > 1 else shapes[0]  # pylint: disable=protected-access
    tess = Tessellator()
    tess.compute(compound, quality, angular_tolerance, compute_faces, compute_edges, debug)
    vertices = tess.get_vertices()
    return {
        "vertices": vertices,
        "triangles": tess.get_triangles(),
        "normals": tess.get_normals(),
        "edges": tess.get_edges(),
    }


def discretize_edge(edge, deflection=0.1):
    curve_adaptator = BRepAdaptor_Curve(edge)

    discretizer = GCPnts_QuasiUniformDeflection()
    discretizer.Initialize(
        curve_adaptator, deflection, curve_adaptator.FirstParameter(), curve_adaptator.LastParameter()
    )

    if not discretizer.IsDone():
        raise AssertionError("Discretizer not done.")

    points = [curve_adaptator.Value(discretizer.Parameter(i)).Coord() for i in range(1, discretizer.NbPoints() + 1)]

    # return tuples representing the single lines of the egde
    edges = []
    for i in range(len(points) - 1):
        edges.append((points[i], points[i + 1]))

    return np.asarray(edges, dtype=np.float32)


def bbox_edges(bb):
    return np.asarray(
        [
            bb["xmax"],
            bb["ymax"],
            bb["zmin"],
            bb["xmax"],
            bb["ymax"],
            bb["zmax"],
            bb["xmax"],
            bb["ymin"],
            bb["zmax"],
            bb["xmax"],
            bb["ymax"],
            bb["zmax"],
            bb["xmax"],
            bb["ymin"],
            bb["zmin"],
            bb["xmax"],
            bb["ymax"],
            bb["zmin"],
            bb["xmax"],
            bb["ymin"],
            bb["zmin"],
            bb["xmax"],
            bb["ymin"],
            bb["zmax"],
            bb["xmin"],
            bb["ymax"],
            bb["zmax"],
            bb["xmax"],
            bb["ymax"],
            bb["zmax"],
            bb["xmin"],
            bb["ymax"],
            bb["zmin"],
            bb["xmax"],
            bb["ymax"],
            bb["zmin"],
            bb["xmin"],
            bb["ymax"],
            bb["zmin"],
            bb["xmin"],
            bb["ymax"],
            bb["zmax"],
            bb["xmin"],
            bb["ymin"],
            bb["zmax"],
            bb["xmax"],
            bb["ymin"],
            bb["zmax"],
            bb["xmin"],
            bb["ymin"],
            bb["zmax"],
            bb["xmin"],
            bb["ymax"],
            bb["zmax"],
            bb["xmin"],
            bb["ymin"],
            bb["zmin"],
            bb["xmax"],
            bb["ymin"],
            bb["zmin"],
            bb["xmin"],
            bb["ymin"],
            bb["zmin"],
            bb["xmin"],
            bb["ymax"],
            bb["zmin"],
            bb["xmin"],
            bb["ymin"],
            bb["zmin"],
            bb["xmin"],
            bb["ymin"],
            bb["zmax"],
        ],
        dtype="float32",
    )



#
# Copyright 2021 Bernhard Walter
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from .utils import warn
from cad_viewer_widget import get_sidecar, get_default_sidecar


class Defaults:
    def __init__(self):
        self.reset_defaults()

    def get_defaults(self):
        return self.defaults

    def get_default(self, key, default_value=None):
        return self.defaults.get(key, default_value)

    def set_defaults(self, **kwargs):
        """Set defaults for CAD viewer

        Valid keywords:

        DISPLAY OPTIONS
        - viewer:             Name of the sidecar viewer (default=""):
                              "" uses the default sidecar (if exists) and None forces to use notebook cell
        - anchor:             How to open sidecar: "right", "split-right", "split-bottom", ... (default="right")
        - cad_width:          Width of CAD view part of the view (default=800)
        - tree_width:         Width of navigation tree part of the view (default=250)
        - height:             Height of the CAD view (default=600)
        - theme:              Theme "light" or "dark" (default="light")
        - pinning:            Allow replacing the CAD View by a canvas screenshot (default=True in cells, else False)

        TESSELLATION OPTIONS
        - angular_tolerance:  Shapes: Angular deflection in radians for tessellation (default=0.2)
        - deviation:          Shapes: Deviation from linear deflection value (default=0.1)
        - edge_accuracy:      Edges: Precision of edge discretization (default=None, i.e. mesh quality / 100)
        - default_color:      Default face color (default=(232, 176, 36))
        - default_edge_color: Default edge color (default="#707070")
        - optimal_bb:         Use optimal bounding box (default=False)
        - render_normals:     Render vertex normals(default=False)
        - render_edges:       Render edges  (default=True)
        - render_mates:       Render mates (for MAssemblies, default=False)
        - mate_scale:         Scale of rendered mates (for MAssemblies, default=1)

        VIEWER OPTIONS
        - control:            Use trackball controls ('trackball') or orbit controls ('orbit') (default='trackball')
        - up:                 Use z-axis ('Z') or y-axis ('Z') as up direction for the camera
        - axes:               Show axes (default=False)
        - axes0:              Show axes at (0,0,0) (default=False)
        - grid:               Show grid (default=[False, False, False])
        - ticks:              Hint for the number of ticks in both directions (default=10)
        - ortho:              Use orthographic projections (default=True)
        - transparent:        Show objects transparent (default=False)
        - black_edges:        Show edges in black (default=False)
        - position:           Absolute camera position that will be scaled (default=None)
        - quaternion:         Camera rotation as quaternion (x, y, z, w) (default=None)
        - target:             Camera target to look at (default=None)
        - zoom:               Zoom factor of view (default=2.5)
        - reset_camera:       Reset camera position, rotation and zoom to default (default=True)
        - zoom_speed:         Mouse zoom speed (default=1.0)
        - pan_speed:          Mouse pan speed (default=1.0)
        - rotate_speed:       Mouse rotate speed (default=1.0)
        - ambient_intensity   Intensity of ambient light (default=0.75)
        - direct_intensity    Intensity of direct lights (default=0.15)
        - show_parent:        Show the parent for edges, faces and vertices objects
        - show_bbox:          Show bounding box (default=False)
        - tools:              Show the viewer tools like the object tree (default=True)
        - glass:              Show the viewer in glass mode, i.e (CAD navigation as transparent overlay (default=False)
        - timeit:             Show rendering times, levels = False, 0,1,2,3,4,5 (default=False)
        - parallel:           (Linux only) Whether to use multiprocessing for parallel tessellation
        - js_debug:           Enable debug output in browser console (default=False)

        NOT SUPPORTED ANY MORE:
        - mac_scrollbar       The default now
        - bb_factor:          Removed
        - display             Use 'viewer="<viewer title>"' (for sidecar display) or 'viewer=None' (for cell display)
        - quality             Use 'deviation'to control smoothness of rendered edges
        """

        for k, v in kwargs.items():
            if self.get_default(k, float("nan")) == float("nan"):
                print(f"Paramater {k} is not a valid argument for show()")

            # elif k == "parallel" and v and platform.system() != "Linux":
            #     warn("parallel=True only works on Linux. Setting parallel=False")
            #     self.defaults[k] = False

            else:
                self.defaults[k] = v

    def reset_defaults(self):
        self.defaults = {
            #
            # display options
            #
            "viewer": "",
            "anchor": "right",
            "cad_width": 800,
            "tree_width": 250,
            "height": 600,
            "theme": "light",
            "pinning": False,
            #
            # render options
            #
            "angular_tolerance": 0.2,
            "deviation": 0.1,
            "edge_accuracy": None,
            "default_color": (232, 176, 36),
            "default_edge_color": "#707070",
            "optimal_bb": False,
            "render_normals": False,
            "render_edges": True,
            "render_mates": False,
            "parallel": False,
            "mate_scale": 1,
            #
            # viewer options
            #
            "control": "trackball",
            "up": "Z",
            "axes": False,
            "axes0": False,
            "grid": [False, False, False],
            "ticks": 10,
            "ortho": True,
            "transparent": False,
            "black_edges": False,
            "ambient_intensity": 0.75,
            "direct_intensity": 0.15,
            "reset_camera": True,
            "show_parent": True,
            "show_bbox": False,
            "position": None,
            "quaternion": None,
            "target": None,
            "zoom": None,
            "zoom_speed": 1.0,
            "pan_speed": 1.0,
            "rotate_speed": 1.0,
            "collapse": 1,
            "tools": True,
            "glass": False,
            "timeit": False,
            "js_debug": False,
        }


def get_defaults():
    return DEFAULTS.get_defaults()


def get_default(key, default_value=None):
    return DEFAULTS.get_default(key, default_value)


def set_defaults(**kwargs):
    if isinstance(kwargs.get("grid"), bool):
        warn(
            "Using bool for grid is deprecated, please use (xy-grid, xz-grid. yz-grid)",
            DeprecationWarning,
            "once",
        )
        kwargs["grid"] = (kwargs["grid"], False, False)

    DEFAULTS.set_defaults(**kwargs)


def apply_defaults(**kwargs):

    viewer = kwargs.get("viewer")
    if viewer is None:
        viewer = get_default_sidecar()
    sidecar = get_sidecar(viewer)

    result = dict(get_defaults())
    for k, v in kwargs.items():
        if result.get(k, float("nan")) != float("nan"):  # use a value that will never be used ("" and None are used)
            result[k] = v
        else:
            print(f"unknown parameter {k}")

    for k in ["anchor", "theme", "pinning"]:
        # omit create args that cannot be set after viewer is created, unless explicit given
        # -> leading to a warning
        if sidecar is not None and kwargs.get(k) is None:
            del result[k]

    return result


def reset_defaults():
    DEFAULTS.reset_defaults()


def create_args(config):
    adapt = lambda key: "title" if key == "viewer" else key

    return {
        adapt(k): v
        for k, v in config.items()
        if k
        in ["viewer", "title", "anchor", "cad_width", "tree_width", "height", "theme", "pinning", "tools", "glass"]
    }


def add_shape_args(config):
    args = {
        k: v
        for k, v in config.items()
        if k
        in [
            "control",
            "up",
            "axes",
            "axes0",
            "grid",
            "ticks",
            "ortho",
            "transparent",
            "black_edges",
            "position",
            "quaternion",
            "target",
            "zoom",
            "reset_camera",
            "ambient_intensity",
            "direct_intensity",
            "default_edge_color",
            "zoom_speed",
            "pan_speed",
            "rotate_speed",
            "clipIntersection",
            "clipPlaneHelpers",
            "clipNormal",
            "collapse",
            "tools",
            "glass",
            "cad_width",
            "tree_width",
            "height",
            "timeit",
            "js_debug",
        ]
    }

    return args


def tessellation_args(config):
    return {
        k: v
        for k, v in config.items()
        if k
        in [
            "angular_tolerance",
            "deviation",
            "edge_accuracy",
            "default_color",
            "default_edge_color",
            "optimal_bb",
            "render_normals",
            "render_edges",
            "render_mates",
            "mate_scale",
            "quality",
            "parallel",
        ]
    }


def show_args(config):
    args = create_args(config)
    args.update(add_shape_args(config))

    if config.get("normal_len") is not None:
        args["normal_len"] = config["normal_len"]
    args["title"] = config.get("viewer")
    return args


def preset(key, value):
    return get_default(key) if value is None else value


DEFAULTS = Defaults()



import os
import pickle
import time
import unicodedata

from OCP.STEPCAFControl import STEPCAFControl_Reader
from OCP.TDF import TDF_LabelSequence, TDF_Label, TDF_ChildIterator
from OCP.TCollection import TCollection_ExtendedString
from OCP.TDocStd import TDocStd_Document
from OCP.XCAFDoc import XCAFDoc_DocumentTool, XCAFDoc_ColorSurf, XCAFDoc_ColorGen, XCAFDoc_ColorCurv
from OCP.TDataStd import TDataStd_Name
from OCP.TCollection import TCollection_AsciiString
from OCP.Quantity import Quantity_ColorRGBA
from OCP.TopLoc import TopLoc_Location
from OCP.TopAbs import TopAbs_SOLID, TopAbs_COMPOUND, TopAbs_COMPSOLID, TopAbs_FACE
from OCP.TopExp import TopExp_Explorer

from .ocp_utils import serialize, deserialize, loc_to_tq, tq_to_loc
from .utils import warn

import cadquery as cq

DEFAULT_COLOR = (0.8, 0.8, 0.8, 1)


def clean_string(s):
    return (
        "".join(ch for ch in s if unicodedata.category(ch)[0] != "C")
        .replace(" ", "_")
        .replace(".", "_")
        .replace("(", "_")
        .replace(")", "_")
    )


class StepReader:
    def __init__(self, analyse_faces=True, split_compounds=True, use_colors=True):
        self.analyse_faces = analyse_faces
        self.split_compounds = split_compounds
        self.use_colors = use_colors
        self.shape_tool = None
        self.color_tool = None
        self.assemblies = None

    def _create_assembly_object(self, name, loc=None, color=None, shape=None, children=None):
        """
        Create a new object
        :param name: object name
        :param loc: object location (TopLoc_Location)
        :param color: 4 tuple (RGBA) with values 0<=x<=1
        :param shape: object shape (TopoDS_Shape)
        :param children: list of AssemblyObject objects
        :return: AssemblyObject
        """
        return {
            "name": name,
            "loc": loc,
            "color": color,
            "shape": shape,
            "shapes": children,
        }

    def get_name(self, label):
        """
        Get name of a TDF_Label object
        :param label: TDF_label of a STEP file
        :return: str
        """
        t = TDataStd_Name()
        if label.FindAttribute(TDataStd_Name.GetID_s(), t):
            name = TCollection_AsciiString(t.Get()).ToCString()
            return clean_string(name)
        else:
            return "Component"

    def get_color(self, shape):
        """
        Get color of a TDF_Label object:
        - if self.use_color is False, return DEFAULT_COLOR, else analyse colors
        - if self.analyse_faces, get all colors of all faces. if all faces have the same color, return it, else return the shape color
        Note: This is BEST EFFORT only. Jupyter-CadQuery does not support different colors for the faces of a solid/compound.
              So for many STEP files with colored faces, the result will not be correct and depend on the structure of the STEP labels
        :param label: TDF_label or TopoDS_Shape of a STEP file
        :return: str
        """

        def to_list(c):
            return (c.GetRGB().Red(), c.GetRGB().Green(), c.GetRGB().Blue(), c.Alpha())

        def get_col(obj):
            col = Quantity_ColorRGBA()
            if (
                self.color_tool.GetColor(obj, XCAFDoc_ColorGen, col)
                or self.color_tool.GetColor(obj, XCAFDoc_ColorSurf, col)
                or self.color_tool.GetColor(obj, XCAFDoc_ColorCurv, col)
            ):
                return to_list(col)

        if not self.use_colors:
            return DEFAULT_COLOR

        shape_color = get_col(shape)

        colors = []
        if self.analyse_faces:

            # Find all face colors
            exp = TopExp_Explorer(shape, TopAbs_FACE)
            while exp.More():
                color = get_col(exp.Current())
                if color is not None:
                    colors.append(color)
                exp.Next()

            colors = list(set(colors))

        # If all faces have the same color, use this as shape color
        if len(colors) == 1:
            return colors[0]
        else:
            return DEFAULT_COLOR if shape_color is None else shape_color

    def get_location(self, label):
        """
        Get location of a TDF_Label object
        :param label: TDF_label of a STEP file
        :return: TopLoc_Location
        """
        return self.shape_tool.GetLocation_s(label)

    def get_shape(self, label):
        """
        Get shape of a TDF_Label object
        :param label: TDF_label of a STEP file
        :return: TopoDS_Shape
        """
        return self.shape_tool.GetShape_s(label)

    def get_shape_details(self, label, name, loc):
        """
        Get shape details of a TopAbs_COMPOUND or TopAbs_COMPSOLID
        :param label: TDF_label of a STEP file
        :param name: object name
        :param loc: object location (TopLoc_Location)
        :return: list of TopAbs_SOLID
        """
        it = TDF_ChildIterator()
        it.Initialize(label)
        i = 0
        shapes = []
        while it.More():
            shape = self.get_shape(it.Value())
            if shape.ShapeType() == TopAbs_SOLID:
                s_name = f"{name}_{i+1}"
                color = self.get_color(shape)
                sub_shape = self._create_assembly_object(s_name, loc, color, shape)
                shapes.append(sub_shape)
                i += 1

            elif shape.ShapeType == TopAbs_COMPSOLID:
                warn(f"Nested compsolids not supported yet: {name}")

            it.Next()

        return shapes

    def get_subshapes(self, label=None, loc=None):
        """
        Get sub shapes of STEP assemblies
        :param label: TDF_label of a STEP file
        :param loc: object location (TopLoc_Location)
        :return: list of AssemblyObjects
        """
        labels = TDF_LabelSequence()
        if label is None:
            # Get all non referenced top level labels
            self.shape_tool.GetFreeShapes(labels)
        else:
            # get all sub-components of the label
            self.shape_tool.GetComponents_s(label, labels)

        result = []

        for i in range(labels.Length()):
            sub_label = labels.Value(i + 1)

            if self.shape_tool.IsReference_s(sub_label):
                ref_label = TDF_Label()
                self.shape_tool.GetReferredShape_s(sub_label, ref_label)
            else:
                ref_label = sub_label

            is_assembly = self.shape_tool.IsAssembly_s(ref_label)

            # Get location from the sub_label and everything else from the referenced label
            loc = self.get_location(sub_label)
            name = self.get_name(ref_label)
            shape = self.get_shape(ref_label)

            sub_shape = self._create_assembly_object(name, loc)

            if is_assembly:
                sub_shape["shapes"] = self.get_subshapes(ref_label)

            elif (
                self.split_compounds
                and shape.ShapeType() in [TopAbs_COMPOUND, TopAbs_COMPSOLID]
                and ref_label.HasChild()
            ):
                sub_shapes = self.get_shape_details(ref_label, name, TopLoc_Location())
                if len(sub_shapes) == 0:
                    sub_shape["shape"] = shape
                    sub_shape["color"] = self.get_color(shape)
                else:
                    sub_shape["shapes"] = sub_shapes

            else:
                sub_shape["shape"] = shape
                sub_shape["color"] = self.get_color(shape)

            result.append(sub_shape)

        return result

    def load(self, filename, cache_name=None, clear_cache=False):
        """
        Load a STEP file
        The result will be stores as a list of AssemblyObjects in self.assemblies and
        for faster reload saved in a pickle format with binary BRep buffers
        :param filename: name of the STEP file
        :param cache_name: name of the binary cache object
        :param clear_cache: clear cache before loading to force analysis of STEP file
        """
        start = time.time()
        if cache_name is not None:
            cache_filename = f"{cache_name}.jq"
            if os.path.exists(cache_filename):
                if clear_cache:
                    os.unlink(cache_filename)
                    print("Cache cleared")
                else:
                    print("Loading from cache ... ", flush=True, end="")
                    self.load_assembly(cache_filename)
                    print("done")
                    print(f"duration: {time.time() - start:5.1f} s")
                    return

        if not os.path.exists(filename):
            raise FileNotFoundError(filename)

        print("Reading STEP file ... ", flush=True, end="")
        time.sleep(0.01)  # ensure output is shown

        fmt = TCollection_ExtendedString("CadQuery-XCAF")
        doc = TDocStd_Document(fmt)

        self.shape_tool = XCAFDoc_DocumentTool.ShapeTool_s(doc.Main())
        self.color_tool = XCAFDoc_DocumentTool.ColorTool_s(doc.Main())

        reader = STEPCAFControl_Reader()
        reader.SetNameMode(True)
        reader.SetColorMode(True)
        reader.SetLayerMode(True)

        reader.ReadFile(filename)
        reader.Transfer(doc)

        print("parsing Assembly ... ", flush=True, end="")

        self.assemblies = self.get_subshapes()

        print("done")
        print(f"duration: {time.time() - start:5.1f} s")

        if cache_name is not None:
            print("Saving to cache ... ", flush=True, end="")
            self.save_assembly(cache_filename)
            print("done")

    def to_cadquery(self, path=None):
        """
        Convert internal AssemblyObjects format to CadQuery Assemblies
        :return: cadquery.Assembly
        """

        def to_workplane(obj):
            return cq.Workplane(obj=cq.Solid(obj))

        def walk(objs, name=None, loc=None):
            a = cq.Assembly(name=name, loc=loc)
            names = {}
            for obj in objs:
                name = obj["name"]

                # Create a unique name by postfixing the enumerator index if needed
                if names.get(name) is None:
                    names[name] = 0
                else:
                    names[name] += 1
                name = f"{obj['name']}_{names[name]}"

                a.add(
                    to_workplane(obj["shape"]) if obj["shapes"] is None else walk(obj["shapes"]),
                    name=name,
                    color=None if obj["color"] is None else cq.Color(*obj["color"]),
                    loc=cq.Location(obj.get("loc")),
                )

            return a

        if len(self.assemblies) == 0 or (
            self.assemblies[0]["shapes"] is not None and len(self.assemblies[0]["shapes"]) == 0
        ):
            raise ValueError("Empty assembly list")

        if len(self.assemblies) == 1:
            assembly = self.assemblies[0]
            return walk(assembly["shapes"], assembly["name"], cq.Location(assembly["loc"]))
        else:
            result = cq.Assembly(name="Group")
            for assembly in self.assemblies:
                result.add(walk(assembly["shapes"], assembly["name"], cq.Location(assembly["loc"])))

        if path != None:
            result = result.objects[path].obj

        return result

    def save_assembly(self, filename):
        """
        Cache the STEP file in a pickle file with binary BRep buffers
        :param filename: name of the cache object
        """

        def _save_assembly(assemblies):
            if assemblies is None:
                return None

            result = []
            for assembly in assemblies:
                obj = self._create_assembly_object(
                    assembly["name"],
                    loc_to_tq(assembly["loc"]),
                    assembly["color"],
                    serialize(assembly["shape"]),
                    _save_assembly(assembly["shapes"]),
                )
                result.append(obj)
            return result

        objs = _save_assembly(self.assemblies)
        with open(filename, "wb") as fd:
            pickle.dump(objs, fd)

    def load_assembly(self, filename):
        """
        Load the STEP file from a pickle file with binary BRep buffers.
        The result will be stores as a list of AssemblyObjects in self.assemblies
        :param filename: name of the cache object
        """

        def _load_assembly(objs):
            if objs is None:
                return None

            result = []
            for obj in objs:
                assembly = self._create_assembly_object(
                    obj["name"],
                    tq_to_loc(*obj["loc"]),
                    obj["color"],
                    deserialize(obj["shape"]),
                    _load_assembly(obj["shapes"]),
                )
                result.append(assembly)
            return result

        with open(filename, "rb") as fd:
            self.assemblies = _load_assembly(pickle.load(fd))



#
# Copyright 2019 Bernhard Walter
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import traceback
from dataclasses import dataclass, field
from typing import Any, List, Dict

from IPython.display import display
from IPython import get_ipython

from ipywidgets import HBox, Output, SelectMultiple, Layout

import cadquery as cq

from .cad_objects import to_assembly, PartGroup, Part, show
from .base import _combined_bb

#
# The Runtime part
# It hooks into __getattribute__ and intercept EVERY python call
# One should only enable replay when necessary for debugging
#


def attributes(names):
    def wrapper(cls):
        for name in names:

            def fget(self, name=name):
                if self.is_empty():
                    raise ValueError("Context empty")
                return self.stack[-1][name]

            def fset(self, value, name=name):
                if self.is_empty():
                    raise ValueError("Context empty")
                self.stack[-1][name] = value

            setattr(cls, name, property(fget, fset))
        return cls

    return wrapper


# pylint: disable=protected-access
@attributes(("func", "args", "kwargs", "obj", "shadow_obj", "children"))
class Context(object):
    def __init__(self):
        self.stack = []
        self.new()
        self.func = None
        self.args = None
        self.kwargs = None
        self.obj = None
        self.shadow_obj = None
        self.children = []

    def _to_dict(self, func, args, kwargs, obj, children, shadow_obj=None):
        return {
            "func": func,
            "args": args,
            "kwargs": kwargs,
            "obj": obj,
            "shadow_obj": shadow_obj,
            "children": children,
        }

    @property
    def length(self):
        return len(self.stack)

    def new(self):
        self.push(None, None, None, None, [], None)

    def clear(self):
        self.stack = []

    def is_empty(self):
        return self.stack == []

    def is_top_level(self):
        return len(self.stack) == 1

    def pop(self):
        if not self.is_empty():
            result = self.stack[-1]
            self.stack = self.stack[:-1]
            return result
        else:
            raise ValueError("Empty context")

    def push(self, func, args, kwargs, obj, children, shadow_obj=None):
        self.stack.append(self._to_dict(func, args, kwargs, obj, children, shadow_obj))

    def append(self, func, args, kwargs, obj, children, shadow_obj=None):
        self.stack[-1].append(self._to_dict(func, args, kwargs, obj, children, shadow_obj))

    def update(self, func, args, kwargs, obj=None, children=None, shadow_obj=None):
        self.func = func
        self.args = args
        self.kwargs = kwargs
        if obj is not None:
            self.obj = obj
        if shadow_obj is not None:
            self.shadow_obj = shadow_obj
        if children is not None:
            self.children = children

    def append_child(self, context):
        self.stack[-1]["children"].append(context)

    def __repr__(self):
        def join(a, b):
            if a == "":
                return b
            elif b == "":
                return a
            else:
                return f"{a}, {b}"

        prefix = "    " * (self.length - 1)
        if self.is_empty():
            result = f"{prefix}   >> Context empty"
        else:
            result = ""
            for i, e in enumerate(self.stack):
                args = "" if e["args"] is None else f'{e["args"]}'[1:-1]
                kwargs = "" if e["kwargs"] is None else ", ".join([f"{k}={v}" for k, v in e["kwargs"].items()])
                result += f'{prefix} >> {i}: {e["func"]}({join(args, kwargs)}) -> {e["obj"]}, {e["shadow_obj"]}/ {e["children"]}\n'
        return result


_CTX = Context()
DEBUG = False
REPLAY = False


def _trace(*objs):
    if DEBUG:
        print(*objs)


def get_context():
    return _CTX


def _add_context(self, name):
    def _blacklist(name):
        return name.startswith("_") or name in (
            "Workplane",
            "val",
            "vals",
            "all",
            "size",
            "add",
            "toOCC",
            "findSolid",
            "findFace",
            "toSvg",
            "exportSvg",
            "largestDimension",
            "Sketch",
            "_edges",
            "_faces",
            "_selection",
            "locs",
        )

    def _is_recursive(func):
        return func in ["union", "cut", "intersect", "placeSketch", "sketch"]

    def _is_recursive_end(func):
        return func in ["union", "cut", "intersect", "placeSketch", "finalize"]

    def intercept(func, prefix):
        def f(*args, **kwargs):
            _trace(prefix, f"Calling {func.__name__}{args + (kwargs,)}")
            _trace(_CTX)

            if _is_recursive_end(func.__name__):
                _ = _CTX.pop()
                _trace(prefix, "--> level down")
                _trace(_CTX)

            if _CTX.args is None:
                _trace(prefix, "Updating")
                _CTX.update(func.__name__, args, kwargs)
                _trace(_CTX)

            result = func(*args, **kwargs)

            if func.__name__ == _CTX.func:
                _CTX.obj = result
                context = _CTX.pop()

                if isinstance(result, cq.Sketch):
                    # Now deep clone the current state of the Sketch object
                    new_obj = cq.Sketch()
                    new_obj._faces = context["obj"]._faces.copy()
                    new_obj._edges = [edge.copy() for edge in context["obj"]._edges]
                    new_obj._selection = [
                        (sel if isinstance(sel, cq.Location) else sel.copy()) for sel in context["obj"]._selection
                    ]
                    new_obj.locs = [loc for loc in context["obj"].locs]
                    context["shadow_obj"] = new_obj

                    # for copy, moved, located, which create a copy of the object, copy the _caller stack
                    if func.__self__ != result:
                        try:
                            result._caller = func.__self__._caller
                        except:  # pylint:disable=bare-except
                            pass

                if _CTX.is_empty():
                    if isinstance(result, cq.Sketch):
                        # to not conflict with overridden __getattribute__ us try...except
                        try:
                            result._caller.append(context)
                        except Exception:  # pylint:disable=bare-except,broad-except
                            result._caller = [context]
                    else:
                        result._caller = context

                    _trace(prefix, f"<== finished {func.__name__}, _caller={result._caller}")
                else:
                    if context["func"] != "sketch":
                        _CTX.append_child(context)
                        _trace("<== added child", context)
                _CTX.new()

            _trace(prefix, "Leaving", func.__name__)
            _trace(_CTX)
            return result

        return f

    attr = object.__getattribute__(self, name)
    if callable(attr):
        prefix = "    " * (_CTX.length - 1)
        if not _blacklist(attr.__name__):
            _trace(prefix, "==> intercepting", attr.__name__)
            if _is_recursive(attr.__name__):
                _trace(prefix, "--> level up")
                _CTX.new()
            _trace(_CTX)

            return intercept(attr, prefix)
    return attr


#
# The UI part
# It evaluates and optimizes what the runtime part had collected
#


@dataclass
class Step:
    level: int = 0
    func: str = ""
    args: List[Any] = field(default_factory=list)
    kwargs: Dict[Any, str] = field(default_factory=dict)
    var: str = ""
    result_name: str = ""
    result_obj: Any = None
    shadow_obj: Any = None

    def clear_func(self):
        self.func = self.args = self.kwargs = ""


class Replay(object):
    def __init__(
        self, deviation, angular_tolerance, edge_accuracy, debug, cad_width, height, sidecar=None, show_result=True,
    ):
        self.debug_output = Output()
        self.deviation = deviation
        self.angular_tolerance = angular_tolerance
        self.edge_accuracy = edge_accuracy
        self.debug = debug
        self.cad_width = cad_width
        self.height = height
        self.sidecar = sidecar
        self.show_result = show_result
        self.reset_camera = True
        self.indexes = []
        self.stack = None
        self.result = None
        self.bbox = None
        self.select_box = None

    def format_steps(self, raw_steps):
        def to_code(step, results):
            def to_name(obj):
                if isinstance(obj, (cq.Workplane, cq.Sketch)):
                    name = results.get(obj, None)
                else:
                    name = str(obj)
                return obj if name is None else name

            if step.func != "":
                if step.func == "newObject":
                    args = ("...",)
                else:
                    args = tuple([to_name(arg) for arg in step.args])
                code = "%s%s%s" % ("| " * step.level, step.func, args)
                code = code[:-2] if len(step.args) == 1 else code[:-1]
                if len(step.args) > 0 and len(step.kwargs) > 0:
                    code += ","
                if step.kwargs != {}:
                    code += ", ".join(["%s=%s" % (k, v) for k, v in step.kwargs.items()])
                code += ")"
                if step.result_name != "":
                    code += " => %s" % step.result_name
            elif step.var != "":
                code = "%s%s" % ("| " * step.level, step.var)
            else:
                code = "ERROR"
            return code

        steps = []
        entries = []
        obj_index = 1

        _trace("\nraw_steps")
        for s in raw_steps:
            _trace(s)

        # Build a lookup table of all Sketch and Workplane result objects
        results = {}
        for step in raw_steps:
            results[step.result_obj] = None
            if step.shadow_obj != None:
                results[step.shadow_obj] = None

        # Check all args whether they are Sketch or Workplane objects
        # and provide a variable name
        for step in raw_steps:
            for arg in step.args:
                if isinstance(arg, (cq.Workplane, cq.Sketch)):
                    results[arg] = "_v%d" % obj_index
                    obj_index += 1

        _trace("\nresults")
        for k, v in results.items():
            _trace(k, v)

        for step in raw_steps:
            if results[step.result_obj] is not None:
                step.result_name = results[step.result_obj]
            steps.append(step)

        _trace("\nsteps")
        for s in steps:
            _trace(s)

        last_level = 1000000

        for step in reversed(steps):
            if step.level < last_level:
                last_level = 1000000
                entries.insert(
                    0, (to_code(step, results), step.result_obj if step.shadow_obj is None else step.shadow_obj)
                )
                if step.var != "":
                    last_level = step.level

        return entries

    def to_array(self, workplane, level=0, result_name=""):
        def walk(caller, level=0, result_name=""):
            stack = [
                Step(
                    level,
                    func=caller["func"],
                    args=caller["args"],
                    kwargs=caller["kwargs"],
                    result_name=result_name,
                    result_obj=caller["obj"],
                    shadow_obj=caller["shadow_obj"],
                )
            ]
            for child in reversed(caller["children"]):
                stack = walk(child, level + 1) + stack
                for arg in child["args"]:
                    if isinstance(arg, cq.Workplane):
                        result_name = getattr(arg, "name", None)
                        stack = self.to_array(arg, level=level + 2, result_name=result_name) + stack
            return stack

        stack = []

        obj = workplane
        while obj is not None:
            caller = getattr(obj, "_caller", None)
            result_name = getattr(obj, "name", "")
            if caller is not None:
                if isinstance(caller, (list, tuple)):
                    stack = [
                        Step(
                            level,
                            func=c["func"],
                            args=c["args"],
                            kwargs=c["kwargs"],
                            result_name=result_name,
                            result_obj=c["obj"],
                            shadow_obj=c["shadow_obj"],
                        )
                        for c in caller
                    ]
                else:
                    stack = walk(caller, level, result_name) + stack
                    for arg in caller["args"]:
                        if isinstance(arg, (cq.Workplane, cq.Sketch)):
                            result_name = getattr(arg, "name", "")
                            stack = self.to_array(arg, level=level + 1, result_name=result_name) + stack

            obj = obj.parent

        _trace("to_array")
        for s in stack:
            _trace(s)

        return stack

    def select_handler(self, change):
        with self.debug_output:
            if change["name"] == "index":
                self.select(change["new"])

    def select(self, indexes):
        self.debug_output.clear_output()
        with self.debug_output:
            self.indexes = indexes
            steps = [(i, self.stack[i][1]) for i in self.indexes]
            try:
                cad_objs = [to_assembly(step[1], name="Step[%02d]" % step[0], show_parent=False) for step in steps]
            except Exception as ex:  # pylint:disable=broad-except
                print(ex)
                traceback.print_exc()

        # Add hidden result to start with final size and allow for comparison

        if self.show_result and (
            isinstance(self.stack[-1][1], cq.Sketch) or not isinstance(self.stack[-1][1].val(), cq.Vector)
        ):
            # result = Part(self.stack[-1][1], "Result", show_faces=False, show_edges=False)
            objs = PartGroup([self.result] + cad_objs, name="Replay")
            show_bbox = False
        else:
            objs = PartGroup(cad_objs, name="Replay")
            show_bbox = self.bbox
        with self.debug_output:
            try:
                show(
                    objs,
                    deviation=self.deviation,
                    angular_tolerance=self.angular_tolerance,
                    edge_accuracy=self.edge_accuracy,
                    reset_camera=self.reset_camera,
                    show_parent=False,
                    show_bbox=show_bbox,
                    viewer=""
                )
                self.reset_camera = False
            except Exception:  # pylint:disable=broad-except
                print("\nWarning: object cannot be shown")


def replay(
    cad_obj,
    index=-1,
    deviation=0.1,
    angular_tolerance=0.2,
    edge_accuracy=None,
    debug=False,
    cad_width=600,
    height=600,
    sidecar=None,
    show_result=False,
):
    while not _CTX.is_top_level:
        _CTX.pop()

    if not REPLAY:
        print("Replay is not enabled. To do so call 'enable_replay()'. Falling back to 'show()'")
        return show(cad_obj, cad_width=cad_width, height=height)
    else:
        print("Use the multi select box below to select one or more steps you want to examine")

    r = Replay(deviation, angular_tolerance, edge_accuracy, debug, cad_width, height, sidecar, show_result,)

    if isinstance(cad_obj, (cq.Workplane, cq.Sketch)):
        workplane = cad_obj
    else:
        print("Cannot replay", cad_obj)
        return None

    r.stack = r.format_steps(r.to_array(workplane, result_name=getattr(workplane, "name", None)))

    # save overall result
    r.result = Part(r.stack[-1][1], "Result", show_faces=True, show_edges=False)

    # tessellate and get bounding box
    shapes = PartGroup([r.result], loc=cq.Location()).collect_shapes(
        "", cq.Location(), deviation=0.1, angular_tolerance=0.2, edge_accuracy=0.01, render_edges=False,
    )
    # save bounding box of overall result
    r.bbox = _combined_bb(shapes).to_dict()

    if index == -1:
        r.indexes = [len(r.stack) - 1]
    else:
        r.indexes = [index]

    r.select_box = SelectMultiple(
        options=["[%02d] %s" % (i, code) for i, (code, obj) in enumerate(r.stack)],
        index=r.indexes,
        rows=len(r.stack),
        description="",
        disabled=False,
        layout=Layout(width="600px"),
    )
    r.select_box.add_class("monospace")
    r.select_box.observe(r.select_handler)
    display(HBox([r.select_box, r.debug_output]))

    r.select(r.indexes)
    return r


#
# Control functions to enable, disable and reset replay
#


def reset_replay():
    _CTX.clear()
    _CTX.new()


def enable_replay(warning=True, debug=False):
    global DEBUG, REPLAY  # pylint:disable=global-statement

    DEBUG = debug

    print("\nEnabling jupyter_cadquery replay")
    cq.Workplane.__getattribute__ = _add_context
    cq.Sketch.__getattribute__ = _add_context

    if warning:
        print("Note: To get rid of this warning, use 'enable_replay(False)'")
        ip = get_ipython()
        if not "reset_replay" in [f.__name__ for f in ip.events.callbacks["pre_run_cell"]]:
            ip.events.register("pre_run_cell", reset_replay)
    REPLAY = True


def disable_replay():
    global REPLAY  # pylint:disable=global-statement
    print("Removing replay from cadquery.Workplane (will show a final RuntimeWarning if not suppressed)")
    cq.Workplane.__getattribute__ = object.__getattribute__

    ip = get_ipython()
    if "reset_replay" in [f.__name__ for f in ip.events.callbacks["pre_run_cell"]]:
        ip.events.unregister("pre_run_cell", reset_replay)
    REPLAY = False



import base64
import base64
from datetime import datetime
from time import localtime
import os
import pickle
import threading
import time
import zmq

from IPython.display import display, clear_output
import ipywidgets as widgets
from jupyter_cadquery import show
from jupyter_cadquery import AnimationTrack
from jupyter_cadquery.defaults import get_default, create_args, add_shape_args, set_defaults
from jupyter_cadquery.logo import LOGO_DATA
from jupyter_cadquery.utils import px

VIEWER = None


def _log(typ, *msg):
    ts = datetime(*localtime()[:6]).isoformat()
    prefix = f"{ts} ({typ}) "
    if VIEWER is not None:
        if isinstance(msg, (tuple, list)):
            VIEWER.log_output.append_stdout(prefix + " ".join([str(m) for m in msg]) + "\n")
        else:
            VIEWER.log_output.append_stdout(prefix + str(msg) + "\n")
    else:
        print(prefix, *msg)


def info(*msg):
    _log("I", *msg)


def warn(*msg):
    _log("W", *msg)


def error(*msg):
    _log("E", *msg)


def debug(*msg):
    _log("D", *msg)


class Viewer:
    def __init__(self, zmq_port):
        self.zmq_port = zmq_port
        self.viewer = None
        self.interactive = None
        self.zmq_server = None
        self.log_output = widgets.Output(layout=widgets.Layout(height="400px", overflow="scroll"))
        self.splash = None
        self.log_view = None

    def _display(self, data, logo=False):
        mesh_data = data["data"]
        config = data["config"]

        if logo:
            config["cad_width"] = get_default("cad_width")
            config["tree_width"] = get_default("tree_width")
            config["height"] = get_default("height")

        if config.get("cad_width") is None:
            config["cad_width"] = get_default("cad_width")
        else:
            if config.get("cad_width") < 640:
                warn("cad_width has to be >= 640, setting to 640")
                config["cad_width"] = 640

        if config.get("height") is None:
            config["height"] = get_default("height")
        else:
            if config.get("height") < 400:
                warn("height has to be >= 400, setting to 400")
                config["height"] = 400

        if config.get("tree_width") is None:
            config["tree_width"] = get_default("tree_width")
        else:
            if config.get("tree_width") < 200:
                warn("tree_width has to be >= 200, setting to 200")
                config["tree_width"] = 200
                
        if config.get("glass") is None:
            config["glass"] = get_default("glass")

        width = config["cad_width"] + 6
        if not config["glass"]:
            width += config["tree_width"]

        if self.interactive is not None:
            self.interactive.layout.width = px(width - 30)
        if self.log_output is not None:
            self.log_output.layout.width = px(width - 30)
        self.log_view.layout.width = px(width)

        # Force reset of camera to not inhereit splash settings for first object
        if self.splash:
            config["reset_camera"] = True
            self.splash = False

        kwargs = add_shape_args(config)
        self.viewer.clear_tracks()
        self.viewer.add_shapes(**mesh_data, **kwargs)
        info(create_args(config))
        info(add_shape_args(config))

    def start_viewer(self, cad_width, cad_height, theme, glass_mode):
        info(f"zmq_port:   {self.zmq_port}")
        info(f"theme:      {theme}")
        info(f"cad_width:  {cad_width}")
        info(f"cad_height: {cad_height}")
        info(f"glass_mode: {glass_mode}")

        set_defaults(theme=theme, cad_width=cad_width, height=cad_height, glass=glass_mode)

        # remove jupyter cadquery start message
        clear_output()

        self.viewer = show(theme=theme, cad_width=cad_width, height=cad_height, glass=glass_mode, pinning=False)
        self.splash = True

        self.log_view = widgets.Accordion(children=[self.log_output])
        self.log_view.set_title(0, "Log")
        self.log_view.selected_index = None
        display(self.log_view)

        stop_viewer()

        context = zmq.Context()
        socket = context.socket(zmq.REP)
        for i in range(5):
            try:
                socket.bind(f"tcp://*:{self.zmq_port}")
                break
            except Exception as ex:
                print(f"{ex}: retrying ... ")
                time.sleep(1)

        self.zmq_server = socket
        info("zmq started\n")

        def return_error(error_msg):
            error(error_msg)
            socket.send_json({"result": "error", "msg": error_msg})

        def return_success(t):
            info(f"duration: {time.time() - t:7.2f}")
            socket.send_json({"result": "success"})

        def msg_handler():
            while True:
                msg = socket.recv()
                try:
                    data = pickle.loads(msg)
                except Exception as ex:
                    return_error(str(ex))
                    continue

                if data.get("type") == "data":
                    try:
                        t = time.time()
                        self._display(data)
                        return_success(t)

                    except Exception as ex:
                        error_msg = f"{type(ex).__name__}: {ex}"
                        return_error(error_msg)

                elif data.get("type") == "animation":
                    try:
                        t = time.time()
                        for track in data["data"]:
                            self.viewer.add_track(AnimationTrack(*track))

                        self.viewer.animate(data["config"]["speed"])
                        return_success(t)

                    except Exception as ex:
                        error_msg = f"{type(ex).__name__}: {ex}"
                        return_error(error_msg)

                else:
                    return_error(f"Wrong message type {data.get('type')}")

        thread = threading.Thread(target=msg_handler)
        thread.setDaemon(True)
        thread.start()

    #        self.viewer.info.add_html("<b>zmq server started</b>")

    def stop_viewer(self):
        if self.zmq_server is not None:
            try:
                self.zmq_server.close()
                info("zmq stopped")
                if self.viewer is not None and self.viewer.info is not None:
                    self.viewer.info.add_html("<b>HTTP zmq stopped</b>")
                self.zmq_server = None
                time.sleep(0.5)
            except Exception as ex:
                error("Exception %s" % ex)


def start_viewer():
    global VIEWER

    zmq_port = 5555 if os.environ.get("ZMQ_PORT") is None else os.environ["ZMQ_PORT"]
    cad_width = get_default("cad_width") if os.environ.get("CAD_WIDTH") is None else int(os.environ["CAD_WIDTH"])
    cad_height = get_default("height") if os.environ.get("CAD_HEIGHT") is None else int(os.environ["CAD_HEIGHT"])
    theme = get_default("theme") if os.environ.get("THEME") is None else os.environ["THEME"]
    glass_mode = get_default("glass") if os.environ.get("GLASS_MODE") is None else (os.environ["GLASS_MODE"] == "1")
    
    VIEWER = Viewer(zmq_port)
    VIEWER.start_viewer(cad_width, cad_height, theme, glass_mode)


def stop_viewer():
    VIEWER.stop_viewer()



from glob import glob
import io
import itertools
import os
import platform
import sys
import tempfile

from cachetools import LRUCache, cached

import numpy as np
from quaternion import rotate_vectors

from cadquery import Compound, Location, Color
from cadquery.occ_impl.shapes import downcast
from .utils import distance
from webcolors import hex_to_rgb

from OCP.TopAbs import (
    TopAbs_EDGE,
    TopAbs_FACE,
)
from OCP.TopoDS import TopoDS_Compound, TopoDS_Shape
from OCP.TopExp import TopExp_Explorer

from OCP.StlAPI import StlAPI_Writer

from OCP.gp import gp_Trsf, gp_Quaternion, gp_Vec
from OCP.TopLoc import TopLoc_Location


# Bounding Box
from OCP.TopoDS import TopoDS_Shape
from OCP.BinTools import BinTools
from OCP.Bnd import Bnd_Box
from OCP.BRep import BRep_Tool
from OCP.BRepBndLib import BRepBndLib
from OCP.BRepMesh import BRepMesh_IncrementalMesh
from OCP.BRepTools import BRepTools
from OCP.BRepGProp import BRepGProp
from OCP.GProp import GProp_GProps


MAX_HASH_KEY = 2147483647

#
# Caching helpers
#


def make_key(objs, loc=None, optimal=False):  # pylint: disable=unused-argument
    # optimal is not used and as such ignored
    if not isinstance(objs, (tuple, list)):
        objs = [objs]

    key = (tuple((s.HashCode(MAX_HASH_KEY) for s in objs)), loc_to_tq(loc))
    return key


def get_size(obj):
    size = sys.getsizeof(obj)
    if isinstance(obj, dict):
        size += sum([get_size(v) + len(k) for k, v in obj.items()])
    elif isinstance(obj, (tuple, list)):
        size += sum([get_size(i) for i in obj])
    return size


cache = LRUCache(maxsize=16 * 1024 * 1024, getsizeof=get_size)

#
# Version
#


def occt_version():
    try:
        lib = glob(f"{os.environ['CONDA_PREFIX']}/lib/libTKBRep.*.*.*")[0]
        return lib.split(".so.")[-1]
    except:
        return "(cannot retrieve Open CASCADE version)"


#
# Bounding Box
#


class BoundingBox(object):
    def __init__(self, obj=None, optimal=False):
        self.optimal = optimal
        if obj is None:
            self.xmin = self.xmax = self.ymin = self.ymax = self.zmin = self.zmax = 0
        elif isinstance(obj, BoundingBox):
            self.xmin = obj.xmin
            self.xmax = obj.xmax
            self.ymin = obj.ymin
            self.ymax = obj.ymax
            self.zmin = obj.zmin
            self.zmax = obj.zmax
        elif isinstance(obj, dict):
            self.xmin = obj["xmin"]
            self.xmax = obj["xmax"]
            self.ymin = obj["ymin"]
            self.ymax = obj["ymax"]
            self.zmin = obj["zmin"]
            self.zmax = obj["zmax"]
        else:
            bbox = self._bounding_box(obj)
            self.xmin, self.xmax, self.ymin, self.ymax, self.zmin, self.zmax = bbox

        self._calc()

    def _center_of_mass(self, obj):
        Properties = GProp_GProps()
        BRepGProp.VolumeProperties_s(obj, Properties)
        com = Properties.CentreOfMass()
        return (com.X(), com.Y(), com.Z())

    def _bounding_box(self, obj, tol=1e-6):
        bbox = Bnd_Box()
        if self.optimal:
            BRepTools.Clean_s(obj)
            BRepBndLib.AddOptimal_s(obj, bbox)
        else:
            BRepBndLib.Add_s(obj, bbox)
        if not bbox.IsVoid():
            values = bbox.Get()
            return (values[0], values[3], values[1], values[4], values[2], values[5])
        else:
            c = self._center_of_mass(obj)
            bb = (c[0] - tol, c[0] + tol, c[1] - tol, c[1] + tol, c[2] - tol, c[2] + tol)
            print("\nVoid Bounding Box", bb)
            return bb

    def _calc(self):
        self.xsize = self.xmax - self.xmin
        self.ysize = self.ymax - self.ymin
        self.zsize = self.zmax - self.zmin
        self.center = (
            self.xmin + self.xsize / 2.0,
            self.ymin + self.ysize / 2.0,
            self.zmin + self.zsize / 2.0,
        )
        self.max = max([abs(x) for x in (self.xmin, self.xmax, self.ymin, self.ymax, self.zmin, self.zmax)])

    def is_empty(self):
        return (
            (abs(self.xmax - self.xmin) < 0.01)
            and (abs(self.ymax - self.ymin) < 0.01)
            and (abs(self.zmax - self.zmin) < 0.01)
        )

    def max_dist_from_center(self):
        return max(
            [
                distance(self.center, v)
                for v in itertools.product((self.xmin, self.xmax), (self.ymin, self.ymax), (self.zmin, self.zmax))
            ]
        )

    def max_dist_from_origin(self):
        return max(
            [
                np.linalg.norm(v)
                for v in itertools.product((self.xmin, self.xmax), (self.ymin, self.ymax), (self.zmin, self.zmax))
            ]
        )

    def update(self, bb, minimize=False):
        lower, upper = (max, min) if minimize else (min, max)

        if isinstance(bb, BoundingBox):
            self.xmin = lower(bb.xmin, self.xmin)
            self.xmax = upper(bb.xmax, self.xmax)
            self.ymin = lower(bb.ymin, self.ymin)
            self.ymax = upper(bb.ymax, self.ymax)
            self.zmin = lower(bb.zmin, self.zmin)
            self.zmax = upper(bb.zmax, self.zmax)
        elif isinstance(bb, dict):
            self.xmin = lower(bb["xmin"], self.xmin)
            self.xmax = upper(bb["xmax"], self.xmax)
            self.ymin = lower(bb["ymin"], self.ymin)
            self.ymax = upper(bb["ymax"], self.ymax)
            self.zmin = lower(bb["zmin"], self.zmin)
            self.zmax = upper(bb["zmax"], self.zmax)
        else:
            raise "Wrong bounding box param"

        self._calc()

    def to_dict(self):
        return {
            "xmin": float(self.xmin),
            "xmax": float(self.xmax),
            "ymin": float(self.ymin),
            "ymax": float(self.ymax),
            "zmin": float(self.zmin),
            "zmax": float(self.zmax),
        }

    def __repr__(self):
        return "{xmin:%.2f, xmax:%.2f, ymin:%.2f, ymax:%.2f, zmin:%.2f, zmax:%.2f}" % (
            self.xmin,
            self.xmax,
            self.ymin,
            self.ymax,
            self.zmin,
            self.zmax,
        )


@cached(cache, key=make_key)
def bounding_box(objs, loc=None, optimal=False):
    if isinstance(objs, (list, tuple)):
        compound = Compound._makeCompound(objs)  # pylint: disable=protected-access
    else:
        compound = objs

    return BoundingBox(compound if loc is None else compound.Moved(loc), optimal=optimal)


def np_bbox(p, t, q):
    if p.size == 0:
        return None

    n_p = p.reshape(-1, 3)
    if t is None and q is None:
        v = n_p
    else:
        n_t = np.asarray(t)
        n_q = np.quaternion(q[-1], *q[:-1])
        v = rotate_vectors([n_q], n_p)[0] + n_t

    bbmin = np.min(v, axis=0)
    bbmax = np.max(v, axis=0)
    return {"xmin": bbmin[0], "xmax": bbmax[0], "ymin": bbmin[1], "ymax": bbmax[1], "zmin": bbmin[2], "zmax": bbmax[2]}


# Export STL


def write_stl_file(compound, filename, tolerance=None, angular_tolerance=None):

    # Remove previous mesh data
    BRepTools.Clean_s(compound)

    mesh = BRepMesh_IncrementalMesh(compound, tolerance, True, angular_tolerance)
    mesh.Perform()

    writer = StlAPI_Writer()

    result = writer.Write(compound, filename)

    # Remove the mesh data again
    BRepTools.Clean_s(compound)
    return result


# OCP serialisation


def serialize(shape):
    if shape is None:
        return None

    if platform.system() == "Darwin":
        with tempfile.NamedTemporaryFile() as tf:
            BinTools.Write_s(shape, tf.name)
            with open(tf.name, "rb") as fd:
                buffer = fd.read()
    else:
        bio = io.BytesIO()
        BinTools.Write_s(shape, bio)
        buffer = bio.getvalue()
    return buffer


def deserialize(buffer):
    if buffer is None:
        return None

    shape = TopoDS_Shape()
    if platform.system() == "Darwin":
        with tempfile.NamedTemporaryFile() as tf:
            with open(tf.name, "wb") as fd:
                fd.write(buffer)
            BinTools.Read_s(shape, tf.name)
    else:
        bio = io.BytesIO(buffer)
        BinTools.Read_s(shape, bio)
    return shape


# OCP types and accessors


def is_compound(topods_shape):
    return isinstance(topods_shape, TopoDS_Compound)


def is_shape(topods_shape):
    return isinstance(topods_shape, TopoDS_Shape)


def _get_topo(shape, topo):
    explorer = TopExp_Explorer(shape, topo)
    hashes = {}
    while explorer.More():
        item = explorer.Current()
        hash_value = item.HashCode(MAX_HASH_KEY)
        if hashes.get(hash_value) is None:
            hashes[hash_value] = True
            yield downcast(item)
        explorer.Next()


def get_faces(shape):
    return _get_topo(shape, TopAbs_FACE)


def get_edges(shape):
    return _get_topo(shape, TopAbs_EDGE)


def get_point(vertex):
    p = BRep_Tool.Pnt_s(vertex)
    return (p.X(), p.Y(), p.Z())


def get_rgb(color):
    if color is None:
        return (176, 176, 176)
    rgb = color.wrapped.GetRGB()
    return (int(255 * rgb.Red()), int(255 * rgb.Green()), int(255 * rgb.Blue()))


def get_rgba(color):
    if color is None:
        return (176, 176, 176, 1.0)
    else:
        rgba = color.toTuple()
        return (int(rgba[0] * 255), int(rgba[1] * 255), int(rgba[2] * 255), rgba[3])


def webcol_to_cq(col):
    color = [c / 255.0 for c in hex_to_rgb(col[:7])]
    alpha = 1.0 if len(col) == 7 else int(col[7:9], 16) / 255
    return Color(*color, alpha)


def tq_to_loc(t, q):
    T = gp_Trsf()
    Q = gp_Quaternion(*q)
    V = gp_Vec(*t)
    T.SetTransformation(Q, V)
    return TopLoc_Location(T)


def loc_to_tq(loc):
    if loc is None:
        return (None, None)

    T = loc.Transformation()
    t = T.TranslationPart()
    q = T.GetRotation()
    return ((t.X(), t.Y(), t.Z()), (q.X(), q.Y(), q.Z(), q.W()))


def wrapped_or_None(obj):
    return None if obj is None else obj.wrapped


def __location__repr__(self):
    f = lambda x: f"{x:8.3f}"
    t, q = loc_to_tq(self.wrapped)
    return f"Location: t=({f(t[0])}, {f(t[1])}, {f(t[2])}), q=({f(q[0])}, {f(q[1])}, {f(q[2])}, {f(q[3])})"


Location.__repr__ = __location__repr__  # type: ignore



#
# Copyright 2021 Bernhard Walter
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from jupyter_cadquery import PartGroup, Part
from jupyter_cadquery.cad_objects import to_assembly
from jupyter_cadquery.base import _tessellate_group, get_normal_len, _combined_bb
from jupyter_cadquery.defaults import get_default, get_defaults, preset


import pickle
import zmq

ZMQ_PORT = 5555
REQUEST_TIMEOUT = 2000
OBJECTS = []


def set_port(port):
    global ZMQ_PORT
    ZMQ_PORT = port


def connect(context):
    endpoint = f"tcp://localhost:{ZMQ_PORT}"
    socket = context.socket(zmq.REQ)
    socket.connect(endpoint)
    return socket


def send(data):
    context = zmq.Context()
    socket = connect(context)

    msg = pickle.dumps(data, 4)
    print(" sending ... ", end="")
    socket.send(msg)

    retries_left = 3
    while True:
        if (socket.poll(REQUEST_TIMEOUT) & zmq.POLLIN) != 0:
            reply = socket.recv_json()

            if reply["result"] == "success":
                print("done")
            else:
                print("\n", reply["msg"])
            break

        retries_left -= 1

        # Socket is confused. Close and remove it.
        socket.setsockopt(zmq.LINGER, 0)
        socket.close()
        if retries_left == 0:
            break

        print("Reconnecting to server…")
        # Create new connection
        socket = connect(context)

        print("Resending ...")
        socket.send(msg)


class Progress:
    def update(self):
        print(".", end="", flush=True)


def _convert(*cad_objs, **kwargs):
    color = kwargs.get("default_color")
    if color is None:
        color = get_default("default_color")

    part_group = to_assembly(
        *cad_objs,
        render_mates=kwargs.get("render_mates"),
        mate_scale=kwargs.get("mate_scale", 1),
        default_color=color,
    )

    if len(part_group.objects) == 1 and isinstance(part_group.objects[0], PartGroup):
        part_group = part_group.objects[0]

    # Do not send defaults for postion, rotation and zoom unless they are set in kwargs
    config = {
        k: v
        for k, v in get_defaults().items()
        if not k in ("position", "rotation", "zoom", "cad_width", "tree_width", "height", "glass")
    }

    for k, v in kwargs.items():
        if v is not None:
            config[k] = v

    shapes, states = _tessellate_group(part_group, kwargs, Progress(), config.get("timeit"))

    config["normal_len"] = get_normal_len(
        preset("render_normals", config.get("render_normals")),
        shapes,
        preset("deviation", config.get("deviation")),
    )

    bb = _combined_bb(shapes).to_dict()
    # add global bounding box
    shapes["bb"] = bb

    data = {
        "data": dict(shapes=shapes, states=states),
        "type": "data",
        "config": config,
        "count": part_group.count_shapes(),
    }
    return data


import numpy as np


def to_array(track):
    return [track.path, track.action, track.times, track.values]


def animate(tracks, speed):
    data = {
        "data": [to_array(track) for track in tracks],
        "type": "animation",
        "config": {"speed": speed},
    }
    send(data)


def show(*cad_objs, **kwargs):
    """Show CAD objects in Jupyter

    Valid keywords:
    - height:            Height of the CAD view (default=600)
    - tree_width:        Width of navigation tree part of the view (default=250)
    - cad_width:         Width of CAD view part of the view (default=800)
    - default_color:     Default mesh color (default=(232, 176, 36))
    - default_edgecolor: Default mesh color (default=(128, 128, 128))
    - render_edges:      Render edges  (default=True)
    - render_normals:    Render normals (default=False)
    - render_mates:      Render mates (for MAssemblies)
    - mate_scale:        Scale of rendered mates (for MAssemblies)
    - deviation:         Shapes: Deviation from linear deflection value (default=0.1)
    - angular_tolerance: Shapes: Angular deflection in radians for tessellation (default=0.2)
    - edge_accuracy:     Edges: Precision of edge discretization (default: mesh quality / 100)
    - optimal_bb:        Use optimal bounding box (default=False)
    - axes:              Show axes (default=False)
    - axes0:             Show axes at (0,0,0) (default=False)
    - grid:              Show grid (default=False)
    - ticks:             Hint for the number of ticks in both directions (default=10)
    - ortho:             Use orthographic projections (default=True)
    - transparent:       Show objects transparent (default=False)
    - ambient_intensity  Intensity of ambient ligth (default=1.0)
    - direct_intensity   Intensity of direct lights (default=0.12)
    - position:          Relative camera position that will be scaled (default=(1, 1, 1))
    - rotation:          z, y and y rotation angles to apply to position vector (default=(0, 0, 0))
    - zoom:              Zoom factor of view (default=2.5)
    - reset_camera:      Reset camera position, rotation and zoom to default (default=True)
    - show_parent:       Show the parent for edges, faces and vertices objects
    - theme:             Theme "light" or "dark" (default="light")
    - tools:             Show the viewer tools like the object tree
    - timeit:            Show rendering times, levels = False, 0,1,2,3,4,5 (default=False)

    NOT SUPPORTED ANY MORE:
    - mac_scrollbar      The default now
    - bb_factor:         Removed
    - display            Use 'viewer="<viewer title>"' (for sidecar display) or 'viewer=None' (for cell display)
    - quality            Use 'deviation'to control smoothness of rendered egdes
    """

    data = _convert(*cad_objs, **kwargs)
    send(data)


def show_object(obj, **kwargs):
    global OBJECTS
    OBJECTS.append(Part(obj, name=f"obj_{len(OBJECTS)}"))
    show(PartGroup(OBJECTS), **kwargs)


def reset():
    global OBJECTS

    OBJECTS = []



from __future__ import print_function
from setuptools import setup, find_packages
from glob import glob
import os

here = os.path.dirname(os.path.abspath(__file__))
is_repo = os.path.exists(os.path.join(here, ".git"))

from distutils import log

log.set_verbosity(log.DEBUG)
log.info("setup.py entered")
log.info("$PATH=%s", os.environ["PATH"])

LONG_DESCRIPTION = "An extension to render cadquery objects in JupyterLab via pythreejs"

setup_args = {
    "name": "jupyter_cadquery",
    "version": "3.2.2",
    "description": "An extension to render cadquery objects in JupyterLab via pythreejs",
    "long_description": LONG_DESCRIPTION,
    "include_package_data": True,
    "python_requires": ">=3.6",
    "install_requires": [
        "webcolors~=1.12",
        "voila~=0.3.5",
        "numpy-quaternion==2022.4.1",
        "cad-viewer-widget~=1.3.5",
        "cachetools~=5.2.0",
    ],
    "extras_require": {
        "dev": {"jupyter-packaging", "cookiecutter", "twine", "bumpversion", "black", "pylint", "pyYaml"},
        "prod": {"cadquery==master"},
    },
    "packages": find_packages(),
    "scripts": ["jcv", "jcv.cmd"],
    "zip_safe": False,
    "author": "Bernhard Walter",
    "author_email": "b_walter@arcor.de",
    "url": "https://github.com/bernhard-42/jupyter-cadquery",
    "keywords": ["ipython", "jupyter", "widgets", "CAD", "cadquery"],
    "classifiers": [
        "Development Status :: 5 - Production/Stable",
        "Framework :: IPython",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Multimedia :: Graphics",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
}

setup(**setup_args)



"""Code to be distributed by Multiprocessing, hence a separate file"""

from multiprocessing import shared_memory

from jupyter_cadquery.ocp_utils import deserialize
from jupyter_cadquery.tessellator import tessellate


def mp_tess(name, deviation, quality, angular_tolerance, compute_faces, compute_edges, debug):
    """This function will be pickled by multiprocessing"""
    sm = shared_memory.SharedMemory(name)
    shape = deserialize(bytes(sm.buf[0 : sm.size]))
    t = tessellate([shape], deviation, quality, angular_tolerance, compute_faces, compute_edges, debug)
    sm.close()
    sm.unlink()
    return (name, t)



#
# Copyright 2019 Bernhard Walter
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from collections import namedtuple
import re

VersionInfo = namedtuple("VersionInfo", ["major", "minor", "patch", "release", "build"])


def get_version(version):
    r = re.compile(r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)\-{0,1}(?P<release>\D*)(?P<build>\d*)")
    major, minor, patch, release, build = r.match(version).groups()
    return VersionInfo(major, minor, patch, release, build)


__version__ = "3.2.2"  # DO NOT EDIT THIS DIRECTLY!  It is managed by bumpversion
__version_info__ = get_version(__version__)



import time


class Progress:
    """Simple, self deleting progress bar"""

    def __init__(self, max_value, tick="\u2014", length=60):
        """Init the progress bar with the max length"""
        self.max = max_value
        self.tick = tick
        self.length = length
        self.step = length / max_value
        self.value = 0
        self.start = time.time()
        self.reset()

    def update(self, step=1):
        """Update progress and delete when 100% is reached"""
        self.value = min(self.value + step, self.max)
        s = int(round(self.step * self.value, 0))
        r = int(self.value / self.max * 100)
        t = time.time() - self.start
        print(f"\r{r:3d}% \u22ee{self.tick * s}{' ' * (self.length - s)}\u22ee ({self.value}/{self.max}) {t:5.2f}s", end="")

    def done(self):
        """Finalize the progress bar"""
        self.value = self.max
        self.update()
        print()

    def reset(self):
        """Reset the progress bar"""
        self.value = 0
        self.update(0)

    def clear(self):
        """Remove the progress bar"""
        # print("\r" + " " * (self.length + 30))
        print()
